# =========================================================================== #
import enum
import secrets
import uuid
from datetime import datetime
from typing import (
    Annotated,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generator,
    List,
    Literal,
    Self,
    Set,
    Tuple,
    Type,
    TypeAlias,
    TypeVar,
    overload,
)

from sqlalchemy import (
    CTE,
    BooleanClauseList,
    Column,
    ColumnElement,
    ColumnExpressionArgument,
    CompoundSelect,
    Enum,
    ForeignKey,
    Select,
    String,
    UniqueConstraint,
    and_,
    func,
    literal_column,
    select,
    true,
    union,
)
from sqlalchemy.dialects import mysql
from sqlalchemy.orm import (
    DeclarativeBase,
    InstrumentedAttribute,
    Mapped,
    Session,
    aliased,
    mapped_column,
    object_session,
    relationship,
)
from sqlalchemy.sql import false

# --------------------------------------------------------------------------- #
from captura import __version__, fields
from captura.err import (
    ErrAccessDocumentGrantBase,
    ErrAccessDocumentGrantInsufficient,
    ErrAccessDocumentPending,
    ErrAccessEvent,
    ErrEventGeneral,
    ErrEventKind,
    ErrEventUndone,
    ErrObjMinSchema,
)
from captura.fields import (
    ChildrenAssignment,
    ChildrenCollection,
    ChildrenDocument,
    ChildrenUser,
    KindEvent,
    KindObject,
    KindRecurse,
    Level,
    LevelStr,
    PendingFrom,
    Plural,
    ResolvableLevel,
)

# CONSTANTS, ENUMS, ETC.
#
# NOTE: These enums will be used throughout the program and a should be used
#       to create the types needed by fastapi and typer.
#

UUIDSplit = Tuple[Set[str], Set[str]]
# NOTE: Indexing is important as it is how the front end will get most data.
#       This is done to keep relationships opaque (by keeping the primary keys
#       out of the users views) and to avoid having to specify multiple primary
#       keys for some resources (e.g. `AssocUserDocument`). See
#
#       .. code::
#
#           https://dev.mysql.com/doc/refman/8.0/en/mysql-indexes.html
#


_uuid = mapped_column(
    String(36),
    default=lambda: str(uuid.uuid4()),
    index=True,
)
_uuid_unique = mapped_column(
    String(36),
    default=lambda: str(uuid.uuid4()),
    index=True,
    unique=True,
)
MappedColumnUUID = Annotated[str, _uuid]
MappedColumnUUIDUnique = Annotated[str, _uuid_unique]

MappedColumnDeleted = Annotated[bool, mapped_column(default=False)]


class KindSelect(str, enum.Enum):
    count = "count"  # type: ignore
    uuids = "uuids"


# =========================================================================== #
# Models and MixinsIndex column sqlalchemy


class Base(DeclarativeBase):
    uuid: Mapped[MappedColumnUUIDUnique]
    deleted: Mapped[MappedColumnDeleted]
    __kind__: ClassVar[KindObject]

    # NOTE: Only `classmethod`s should need session due to `object_session`.

    # ----------------------------------------------------------------------- #
    # Finders.
    def q_events(self):
        return (
            select(Event).where(Event.uuid_obj == self.uuid).order_by(Event.timestamp)
        )

    @classmethod
    def if_exists(
        cls,
        session: Session,
        uuid: str,
        # status: int = 404, msg=None
    ) -> Self:
        m = session.execute(select(cls).where(cls.uuid == uuid)).scalar()
        if m is None:
            err = ErrObjMinSchema.httpexception(
                "_msg_dne", 404, uuid_obj=uuid, kind_obj=KindObject(cls.__tablename__)
            )
            raise err
        return m

    @classmethod
    def if_many(
        cls,
        session: Session,
        uuid: Set[str],
        callback: Callable[[Self], Self] | None = None,
    ) -> Tuple[Self, ...]:
        res = session.execute(select(cls).where(cls.uuid.in_(uuid))).scalars()
        if callback is not None:
            return tuple(callback(item) for item in res)

        return tuple(res)

    # ----------------------------------------------------------------------- #
    # Resolvers

    @overload
    @classmethod
    def resolve(cls, session: Session, that: Self | str) -> Self: ...

    @overload
    @classmethod
    def resolve(
        cls, session: Session, that: Tuple[Self, ...] | Set[str]
    ) -> Tuple[Self, ...]: ...

    @classmethod
    def resolve(
        cls, session: Session, that: Self | str | Tuple[Self, ...] | Set[str]
    ) -> Self | Tuple[Self, ...]:
        """Provided :param:`that` which is any reasonable representation of
        a(n) instance(s), return the associated instance(s).
        """
        match that:
            case tuple():
                return that
            case cls():  # type: ignore
                return that
            case str():
                return cls.if_exists(session, that)
            case set():
                return cls.if_many(session, that)
            case _ as bad:
                msg = f"Cannot resolve `{bad}` (of type `{bad}`) for mapped "
                msg += f"class `{cls.__name__}`."
                raise ValueError(msg)

    @overload
    @classmethod
    def resolve_uuid(
        cls,
        session: Session,
        that: Set[str] | Tuple[Self, ...],
    ) -> Set[str]: ...

    @overload
    @classmethod
    def resolve_uuid(
        cls,
        session: Session,
        that: str | Self,
    ) -> str: ...

    @classmethod
    def resolve_uuid(
        cls,
        session: Session,
        that: str | Self | Set[str] | Tuple[Self, ...],
    ) -> str | Set[str]:
        data: Self | Tuple[Self, ...]
        match data := cls.resolve(session, that):
            case cls():  # type: ignore
                return data.uuid
            case tuple():
                return set(item.uuid for item in data)
            case set() | str():
                return data
            case _ as bad:
                raise ValueError(f"Invalid identifier `{bad}`.")

    @classmethod
    def resolve_selectable(
        cls, kind: KindSelect | None = None
    ) -> (
        Type[Self]
        | InstrumentedAttribute[MappedColumnUUID]
        | ColumnExpressionArgument[int]
    ):
        match kind:
            case KindSelect.count:
                return func.count(cls.uuid)
            case KindSelect.uuids:
                return cls.uuid
            case None:
                return cls

        raise ValueError(f"Cannot resolve selectable `{kind}`.")

    # ----------------------------------------------------------------------- #
    # Rest

    @classmethod
    def q_uuid(cls, uuids: set[str]):
        return select(cls).where(cls.uuid.in_(uuids))

    def get_session(self) -> Session:
        session = object_session(self)
        if session is None:
            detail = dict(
                uuid=self.uuid,
                kind=KindObject._value2member_map_[self.__tablename__].name,
                msg="Could not find session.",
            )
            raise ValueError(str(detail))
        return session

    def check_not_deleted(
        self,
        # status_code: int = 410,
        # deleted: bool = True,
    ) -> Self:
        if self.deleted:
            err = ErrObjMinSchema.httpexception(
                "_msg_deleted",
                410,
                uuid_obj=self.uuid,
                kind_obj=KindObject(self.__tablename__),
            )
            raise err
        return self


class PrimaryTableMixins:
    """Creation and deletion data will go into the table associated with
    :class:`Event`.
    """

    public: Mapped[bool] = mapped_column(default=True)
    # NOTE: Now on ``Base``.
    # uuid: Mapped[MappedColumnUUID]
    # deleted: Mapped[bool] = mapped_column(default=False)

    @classmethod
    def q_select_ids(cls, uuids: Set[str]) -> Select:
        match id := getattr(cls, "id", None):
            case Column() | InstrumentedAttribute():
                return select(id).where(cls.uuid.in_(uuids))  # type: ignore[attr-defined]
            case None:
                raise AttributeError(
                    f"Table `{cls.__name__}` must have an `id` `column` to use"
                    "`q_select_ids`."
                )
            case _:
                tt = type(id)
                raise ValueError(f"`id` must be of type `Column`, got `{tt}`.")

    @classmethod
    def q_conds(
        cls,
        uuids: Set[str] | None,
        exclude_deleted: bool = True,
        *,
        conds: BooleanClauseList | ColumnElement[bool] | None = None,
    ):
        items = [conds] if conds is not None else []
        if exclude_deleted:
            items.append(cls.deleted == false())  # type: ignore[attr-defined]
        if uuids is not None:
            items.append(cls.uuid.in_(uuids))  # type: ignore[attr-defined]

        # NOTE: ``true`` is added here because of the following warning:
        #
        #       .. code:: txt
        #
        #          SADeprecationWarning: Invoking and_() without arguments is
        #          deprecated, and will be disallowed in a future release. For
        #          an empty and_() construct, use 'and_(true(), *args)' or
        #          'and_(True, *args)'.
        return and_(true(), *items)  # if len(items) else None

    @classmethod
    def q_conds_public(
        cls,
        uuids: Set[str] | None = None,
        exclude_deleted: bool = True,
    ) -> ColumnElement[bool]:
        conds = cls.public == true()
        return cls.q_conds(uuids, exclude_deleted, conds=conds)  # type: ignore


class SearchableTableMixins(PrimaryTableMixins):
    name: Mapped[str]
    description: Mapped[str]

    @classmethod
    def q_select_public(
        cls, uuids: Set[str] | None = None, exclude_deleted: bool = True
    ):
        q = select(cls).where(cls.q_conds_public(uuids, exclude_deleted))
        return q

    @classmethod
    def q_search_conds(
        cls,
        name_like: str | None = None,
        description_like: str | None = None,
        conds=None,
    ):
        fmt = "^.*{}.*$"
        items = [conds] if conds is not None else []
        if name_like is not None:
            items.append(cls.name.regexp_match(fmt.format(name_like)))
        if description_like is not None:
            items.append(cls.description.regexp_match(fmt.format(description_like)))
        return and_(*items)

    @classmethod
    def q_search(
        cls,
        user_uuid: str | None,
        uuids: Set[str] | None = None,
        exclude_deleted: bool = True,
        *,
        all_: bool = True,
        name_like: str | None = None,
        description_like: str | None = None,
        limit: int | None = None,
        randomize: bool = False,
    ):
        if not all_ and user_uuid is None:
            msg = "`all_` must be true when a user is not provided."
            raise ValueError(msg)

        # TODO: Implement this once it is needed.
        q = None
        if user_uuid:
            # raise ValueError("Not implemented!")
            q = cls.q_select_for_user(
                user_uuid,
                uuids,
                exclude_deleted=exclude_deleted,
            )

        r = None
        if all_:
            r = cls.q_select_public(uuids, exclude_deleted)

        search_conds = cls.q_search_conds(name_like, description_like)
        _items = (
            p.where(search_conds) if search_conds is not None else p
            for p in (q, r)
            if p is not None
        )
        if limit:
            _items = (p.limit(limit) for p in _items)
            if randomize:
                _items = (p.order_by(func.random()) for p in _items)

        items = tuple(_items)
        if not len(items):
            raise ValueError()

        # NOTE: The following steps are necessary to not end up with a
        #       cartesian product in the results. See
        #
        #       .. code:: txt
        #
        #         https://docs.sqlalchemy.org/en/20/orm/queryguide/select.html#selecting-entities-from-unions-and-other-set-operations
        #

        rr = union(*items).subquery()
        cls_alias = aliased(cls, rr)  # type: ignore
        q = select(cls_alias)  # type: ignore
        if limit is not None:
            q = q.limit(limit)
        return q

    @classmethod
    def q_select_for_user(  # type: ignore[empty-body]
        cls,
        user_uuid: str,
        uuids: Set[str] | None = None,
        *,
        exclude_deleted: bool = True,
    ) -> Select: ...


# =========================================================================== #
# Mapped


class Event(Base):
    # NOTE: Would it make more sence to record updates, creations, and
    #       deletions inside of an events table instead? It would be easy to
    #       track the objects by their UUID

    detail_document_editted: ClassVar[str] = "Document editted."

    __tablename__ = "events"
    __kind__ = KindObject.event

    timestamp: Mapped[int] = mapped_column(
        default=(_now := lambda: datetime.timestamp(datetime.now())),
    )

    # id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    uuid: Mapped[MappedColumnUUIDUnique] = mapped_column(primary_key=True)
    uuid_parent: Mapped[str] = mapped_column(
        ForeignKey("events.uuid", ondelete="CASCADE"),
        nullable=True,
    )
    children: Mapped[List["Event"]] = relationship(
        "Event",
        foreign_keys=uuid_parent,
        cascade="delete, all",
    )

    uuid_undo: Mapped[str | None] = mapped_column(
        ForeignKey("events.uuid"),
    )
    undo: Mapped["Event"] = relationship(
        "Event",
        foreign_keys=uuid_undo,
        cascade="delete, all",
    )

    uuid_user: Mapped[str] = mapped_column(ForeignKey("users.uuid", ondelete="CASCADE"))
    user: Mapped["User"] = relationship()

    uuid_obj: Mapped[MappedColumnUUID | None]
    api_origin: Mapped[str] = mapped_column(String(128))
    api_version: Mapped[str] = mapped_column(String(16), default=__version__)
    kind: Mapped[KindEvent] = mapped_column(Enum(KindEvent))
    kind_obj: Mapped[KindObject] = mapped_column(Enum(KindObject))
    detail: Mapped[str | None] = mapped_column(
        String(fields.LENGTH_DESCRIPTION), nullable=True
    )

    def update(
        self,
        session: Session | None = None,
        _recurance: int = 0,
        *,
        api_origin: str | None = None,
        detail: str | None = None,
        uuid_user: str | None = None,
    ):
        "Update safe fields recursively. DOES NOT commit."

        session = session or self.get_session()
        if api_origin is not None:
            self.api_origin = api_origin
        if detail is not None:
            self.detail = detail
        if uuid_user is not None:
            self.uuid_user = uuid_user
        for child in self.children:
            child.update(
                session,
                _recurance + 1,
                api_origin=api_origin,
                detail=detail,
                uuid_user=uuid_user,
            )

    def flattened(self) -> Generator["Event", None, None]:
        yield self
        yield from (
            child_child for child in self.children for child_child in child.flattened()
        )

    @property
    def object_(self) -> "AnyModel | None":
        session = self.get_session()
        t = Tables[self.kind_obj.value].value
        q = select(t).where(t.uuid == self.uuid_obj)
        res = session.execute(q).scalar()  # type: ignore
        if res is not None and res.__tablename__ == self.__tablename__:
            return None
        return res  # type: ignore

    @classmethod
    def cte_recursive(cls, uuid_event: str) -> CTE:
        rand = secrets.token_urlsafe(4)
        uuid_root: Any
        q = (
            select(
                Event,
                uuid_root := literal_column(f"'{uuid_event}'").label("uuid_root"),
                func.cast(Event.uuid, String(512)).label("path"),
                literal_column("0").label("level"),
            )
            .where(Event.uuid == uuid_event)
            .cte(f"find_children_{rand}", recursive=True)
        )

        roots_alias = q.alias(f"BB_{rand}")
        events_alias = Event.__table__.alias(f"AA_{rand}")

        p = select(
            events_alias,
            uuid_root,
            func.concat(roots_alias.c.path, ",", events_alias.c.uuid).label("path"),
            literal_column("level + 1").label("level"),
        ).where(
            events_alias.c.uuid_parent == roots_alias.c.uuid,
            func.find_in_set(events_alias.c.uuid, roots_alias.c.path) == 0,
        )
        q = q.union_all(p)
        return q

    @classmethod
    def q_select_recursive(
        cls,
        uuid_event: Set[str],
        kind_recurse: KindRecurse = KindRecurse.depth_first,
    ) -> CompoundSelect:
        def qq(uuid):
            q = select(cls.cte_recursive(uuid))
            match kind_recurse:
                case KindRecurse.depth_first:
                    q = q.order_by(literal_column("path"))
                case _:
                    pass
            q = q.order_by("timestamp")
            return q

        return union(*(qq(uuid) for uuid in uuid_event))

    @classmethod
    def q_conds(
        cls,
        uuid_user: str | None = None,
        uuid_event: Set[str] | None = None,
        kind: str | None = None,
        kind_obj: str | None = None,
        uuid_obj: str | None = None,
        before: int | None = None,
        after: int | None = None,
    ):
        conds = []
        if uuid_user is not None:
            conds.append(cls.uuid_user == uuid_user)
        if uuid_event is not None:
            conds.append(cls.uuid.in_(uuid_event))
        if kind is not None:
            conds.append(cls.kind == kind)
        if kind_obj is not None:
            conds.append(cls.kind_obj == kind_obj)
        if uuid_obj is not None:
            conds.append(cls.uuid_obj == uuid_obj)
        if before is not None:
            conds.append(cls.timestamp < before)
        if after is not None:
            conds.append(cls.timestamp > after)
        return and_(*conds)

    @classmethod
    def q_select_search(
        cls,
        # From `EventSearchSchema`.
        uuid_user: str | None = None,
        uuid_event: Set[str] | None = None,
        kind: str | None = None,
        kind_obj: str | None = None,
        uuid_obj: str | None = None,
        after: int | None = None,
        before: int | None = None,
        limit: int | None = None,
    ) -> Select:
        q = (
            select(cls)
            .where(
                cls.q_conds(
                    uuid_user=uuid_user,
                    uuid_event=uuid_event,
                    kind=kind,
                    kind_obj=kind_obj,
                    uuid_obj=uuid_obj,
                    before=before,
                    after=after,
                )
            )
            .order_by(cls.timestamp)
        )
        if limit:
            q = q.limit(limit)

        return q

    def check_kind(
        self, kind: KindEvent | None = None, kind_obj: KindObject | None = None
    ) -> Self:
        if kind is not None and self.kind != kind:
            raise ErrEventKind.httpexception(
                "_msg_kind_obj",
                400,
                kind_event=self.kind,
                kind_expected=kind,
                uuid_event=self.uuid,
            )
        if kind_obj is not None and self.kind_obj != kind_obj:
            raise ErrEventKind.httpexception(
                "_msg_kind_event",
                400,
                kind_obj_event=self.kind_obj,
                kind_obj_expected=kind_obj,
                uuid_event=self.uuid,
            )
        return self

    def check_not_undone(self):
        if self.uuid_undo is not None:
            err = ErrEventUndone.httpexception(
                "_msg_undone",
                400,
                uuid_event=self.uuid,
                uuid_event_undo=self.uuid_undo,
            )
            raise err
        return self

    def find_root(self, session: Session | None = None) -> "Event":
        if self.uuid_parent is None:
            return self

        session = session or self.get_session()
        next_ = session.execute(
            select(Event).where(Event.uuid == self.uuid_parent),
        ).scalar()
        if next_ is None:
            raise ErrEventGeneral.httpexception(
                "_msg_undone",
                400,
                uuid_event=self.uuid,
            )

        return next_


class AssocCollectionDocument(Base):
    __tablename__ = "assignments"
    __kind__ = KindObject.assignment

    # NOTE: Since this object supports soft deletion (for the deletion grace
    #       period that will later be implemented) deleted is included.
    # deleted: Mapped[MappedColumnDeleted]
    id_document: Mapped[int] = mapped_column(
        ForeignKey(
            "documents.id",
            ondelete="CASCADE",
        ),
        primary_key=True,
    )

    id_collection: Mapped[int] = mapped_column(
        ForeignKey(
            "collections.id",
            ondelete="CASCADE",
        ),
        primary_key=True,
    )

    @property
    def uuid_document(self) -> str:
        session = self.get_session()
        res = session.execute(
            select(Document.uuid).where(Document.id == self.id_document)
        ).scalar()
        if res is None:
            raise ValueError("Inconcievable!")
        return res

    @property
    def uuid_collection(self) -> str:
        session = self.get_session()
        res = session.execute(
            select(Collection.uuid).where(Collection.id == self.id_collection)
        ).scalar()
        if res is None:
            raise ValueError("Inconcievable!")
        return res

    @classmethod
    def resolve_target_kind(
        cls,
        source: "Collection | Document",
        resolvable_target: "ResolvableMultiple[Collection] | ResolvableMultiple[Document]",
    ) -> ChildrenAssignment:
        match [source, resolvable_target]:
            case [Document()]:
                return ChildrenAssignment.collections
            case [Collection()]:
                return ChildrenAssignment.documents
            case _:
                msg = f"Invalid source `{source}` of type `{type(source)}`. "
                msg += "Must be an instnce of `Collection` or `Document`."
                raise ValueError(msg)


# NOTE: Should be able to be passed directly into `GrantSchema`.
class AssocUserDocument(Base):
    """Also known as :class:`Grant`. Determines user access to documents and
    facilitates invitations.

    See ``Granting Process`` in ``README.md``.

    :attr uuid: Global identifier.
    :attr deleted: Is this grant pending deletion? Grants pending deletion
        that have an active child grant will not be deleted to preserve
        integrity of the table, but will be set in a deleted state. The
        :class:`Event` corresponding to this deletion (generated by
        :class:`Delete`) will be handled by a cronjob or cue that will cleanup
        deleted grants that are not required for table integrity.
    :attr id_user: The user to whom the grant is.
    :attr id_document: The document for which the grant is.
    :attr level: The level provided by the grant, see :class:`Level`.
    :attr pending: Has this grant been accepting by the invited/requested
        party? The grant will not used to verify permissions when accessing
        documents unless pending is false.
    :attr pending_from: How was this grant created. For instance, this field
        should have ``PendingFrom.created`` when it from posting a new
        document, ``PendingFrom.grantee`` when a grantee requests access,
        and ``PendingFrom.granter`` when the grantee is invited.
    :attr uuid_parent: This allows the origin of any particular grant to be
        passed back. While a foreign key to ``User`` could be used this breaks
        joins:

        .. code:: STDERR

            sqlalchemy.exc.AmbiguousForeignKeysError: Can't determine join
            between 'users' and 'grants'; tables have more
            than one foreign key constraint relationship between them. Please
            specify the 'onclause' of this join explicitly.

        Further this lacks the ability to trace back the granting history.
    """

    __tablename__ = "grants"
    __kind__ = KindObject.grant

    # https://stackoverflow.com/questions/28843254/deleting-from-self-referential-inherited-objects-does-not-cascade-in-sqlalchemy
    uuid: Mapped[MappedColumnUUIDUnique] = mapped_column(primary_key=True)
    uuid_parent: Mapped[str] = mapped_column(
        ForeignKey(
            "grants.uuid",
            ondelete="CASCADE",
        ),
        nullable=True,
    )
    children: Mapped[List["AssocUserDocument"]] = relationship(
        "AssocUserDocument",
        foreign_keys=uuid_parent,
        cascade="all, delete",
    )

    id_user: Mapped[int] = mapped_column(
        ForeignKey(
            "users.id",
            ondelete="CASCADE",
        ),
        key="a",
    )
    id_document: Mapped[int] = mapped_column(
        ForeignKey(
            "documents.id",
            ondelete="CASCADE",
        ),
        key="b",
    )
    level: Mapped[Level] = mapped_column(Enum(Level))

    # Metadata
    pending: Mapped[bool] = mapped_column(default=True)
    pending_from: Mapped[PendingFrom] = mapped_column(Enum(PendingFrom))

    __table_args__ = (UniqueConstraint("a", "b", name="_grant_vector"),)

    @property
    def uuid_document(self) -> str:
        session = self.get_session()
        res = session.execute(
            select(Document.uuid).where(Document.id == self.id_document)
        ).scalar()
        if res is None:
            raise ValueError("Inconcievable!")
        return res

    @property
    def uuid_user(self) -> str:
        session = self.get_session()
        res = session.execute(select(User.uuid).where(User.id == self.id_user)).scalar()
        if res is None:
            raise ValueError("Inconcievable!")
        return res

    @property
    def uuid_user_granter(self) -> str:
        session = self.get_session()
        res = session.execute(
            select(User.uuid).where(User.id == self.id_user_granter)  # type: ignore
        ).scalar()
        if res is None:
            raise ValueError("Inconcievable!")
        return res

    @classmethod
    def resolve_from_target(
        cls, session: Session, source: "User | Document", uuid_target: Set[str]
    ) -> Tuple[Self, ...]:
        q = select(cls).join(User).join(Document)
        match source:
            case User(uuid=uuid_user):
                conds = (User.uuid == uuid_user, Document.uuid.in_(uuid_target))
            case Document(uuid=uuid_doc):
                conds = (Document.uuid == uuid_doc, User.uuid.in_(uuid_target))
            case bad:
                raise ValueError(f"Invalid input `{bad}`.")

        q = q.where(*conds)
        return tuple(session.execute(q).scalars())

    # ----------------------------------------------------------------------- #
    # Eventually to merged with the corresponding methods of `Assignment`

    @classmethod
    def q_split(
        cls,
        source: "User | Document",
        target_uuids: Set[str],
        *,
        select_parent_uuids: bool = False,
    ) -> Tuple[Select, Select]:
        "Returns a select for assignment uuids."

        # NOTE: These queries should also have joins such that warning are not
        #       raised by sqlalchemy.
        s: Any
        match source:
            case _ if not select_parent_uuids:
                q = source.q_select_grants(target_uuids, exclude_deleted=False)
                s = literal_column("uuid")
            case User() as user:
                q = user.q_select_documents(target_uuids, exclude_deleted=False)
                s = literal_column("uuid_user")
            case Document() as document:
                q = document.q_select_users(target_uuids, exclude_deleted=False)
                s = literal_column("uuid_document")

        return tuple(  # type: ignore
            select(s).select_from(q.where(Grant.deleted == bool_()).subquery())
            for bool_ in (true, false)
        )


class User(SearchableTableMixins, Base):
    __tablename__ = "users"
    __kind__ = KindObject.user

    id: Mapped[int] = mapped_column(
        primary_key=True,
        autoincrement=True,
    )

    # NOTE: subject should be a sha256 of a token subject. For test tokens,
    #       the subject should be the sha sum of their uuid.
    subject: Mapped[str | None] = mapped_column(String(64), unique=True, nullable=True)

    email: Mapped[str] = mapped_column(String(fields.LENGTH_NAME), unique=True)
    name: Mapped[str] = mapped_column(String(fields.LENGTH_NAME))
    description: Mapped[str] = mapped_column(String(fields.LENGTH_DESCRIPTION))
    url_image: Mapped[str] = mapped_column(String(fields.LENGTH_URL), nullable=True)
    url: Mapped[str] = mapped_column(String(fields.LENGTH_URL), nullable=True)
    admin: Mapped[bool] = mapped_column(default=False)
    content: Mapped[Dict[str, Any]] = mapped_column(
        mysql.JSON, default=None, nullable=True
    )

    # NOTE: This should correspond to `user`. Is all of the collection objects
    #       labeled by their names.
    collections: Mapped[List["Collection"]] = relationship(
        cascade="all, delete",
        back_populates="user",
        primaryjoin="User.id==Collection.id_user",
        passive_deletes=True,
    )

    documents: Mapped[List["Document"]] = relationship(
        secondary=AssocUserDocument.__table__,
        back_populates="users",
        # NOTE: ``cascade`` is enabled on the database side, not here. Enabling
        #       this breaks things.
        single_parent=True,
    )
    events: Mapped[Event] = relationship(back_populates="user", cascade="all, delete")

    # ----------------------------------------------------------------------- #
    # Queries

    # @classmethod
    # def _q_prototype_activation_pending_approval(
    #     cls,
    #     invitation_uuid: Set[str] | None = None,
    #     invitation_email: Set[str] | None = None,
    #     invitation_code: Set[str] | None = None,
    # ) -> Select:
    #     q = select(cls).where(
    #         cls.deleted == true(), cls._prototype_activation_pending_approval == true()
    #     )
    #     if invitation_uuid is not None:
    #         q = q.where(cls.uuid.in_(invitation_uuid))
    #
    #     if invitation_code is not None:
    #         q = q.where(cls._prototype_activation_invitation_code.in_(invitation_code))
    #
    #     if invitation_email is not None:
    #         q = q.where(
    #             cls._prototype_activation_invitation_email.in_(invitation_email)
    #         )
    #
    #     return q

    def q_conds_grants(
        self,
        document_uuids: None | Set[str] = None,
        level: ResolvableLevel | None = None,
        level_levelsets: bool = False,
        exclude_deleted: bool = True,
        pending: bool | None = None,
        pending_from: PendingFrom | None = None,
        exclude_pending: bool = True,
        n_owners: int | None = None,
    ):
        cond = list()
        if n_owners is None:
            cond.append(AssocUserDocument.id_user == self.id)

        if exclude_deleted:
            cond.append(Grant.deleted == false())
            cond.append(Document.deleted == false())

        if document_uuids is not None:
            q_ids = select(Document.id)
            q_ids = q_ids.where(Document.uuid.in_(document_uuids))
            cond.append(AssocUserDocument.id_document.in_(q_ids))

        if level is not None:
            level = Level.resolve(level)
            cond.append(
                (Grant.level >= level.value)
                if not level_levelsets
                else (Grant.level == level)
            )

        match (pending, exclude_pending):
            case (True, True):
                msg = "`pending` and `exclude_pending` cannot both be `True`."
                raise ValueError(msg)
            case (bool() as pending, False):
                cond.append(Grant.pending == (true() if pending else false()))
            case (None | False, True):
                cond.append(Grant.pending == false())
            case (None, False):
                pass
            case bad:
                raise ValueError(f"Cannot handle case `{bad}`.")

        if pending_from is not None:
            cond.append(Grant.pending_from == pending_from)

        return and_(true(), *cond)

    def q_select_grants(
        self,
        document_uuids: None | Set[str] = None,
        level: ResolvableLevel | None = None,
        level_levelsets: bool = False,
        exclude_deleted: bool = True,
        pending: bool | None = None,
        exclude_pending: bool = True,
        pending_from: PendingFrom | None = None,
    ) -> Select:
        # NOTE: Attempting to make roughly the following query:
        #
        #       .. code::
        #
        #          SELECT users.uuid,
        #                 documents.uuid,
        #                 _assocs_user_documents.level
        #          FROM users
        #          JOIN _assocs_user_documents
        #               ON _assocs_user_documents.id_user=users.id
        #          JOIN documents
        #               ON _assocs_user_documents.id_document = documents.id;
        q = select(Grant).select_from(User).join(AssocUserDocument).join(Document)
        conds = self.q_conds_grants(
            document_uuids,
            level,
            level_levelsets=level_levelsets,
            exclude_deleted=exclude_deleted,
            pending=pending,
            exclude_pending=exclude_pending,
            pending_from=pending_from,
        )
        q = q.where(conds)
        return q

    def q_select_documents(
        self,
        document_uuids: Set[str] | None = None,
        level: ResolvableLevel | None = None,
        *,
        level_levelsets: bool = False,
        exclude_deleted: bool = True,
        exclude_pending: bool = True,
        pending: bool = False,
        pending_from: PendingFrom | None = None,
        n_owners: int | None = None,
        n_owners_levelsets: bool = True,
        kind_select: KindSelect | None = None,
    ) -> "Select[Tuple[Document, ...]] | Select[Tuple[int]] | Select[Tuple[str, ...]]":
        """Dual to :meth:`q_select_user_uuids`."""

        selected: Any = Document.resolve_selectable(kind_select)
        q = (
            select(selected)
            .join(Grant)
            .where(
                self.q_conds_grants(
                    document_uuids,
                    level=level,
                    level_levelsets=level_levelsets,
                    exclude_deleted=exclude_deleted,
                    exclude_pending=exclude_pending,
                    pending=pending,
                    pending_from=pending_from,
                    n_owners=n_owners,
                ),
            )
        )

        if n_owners is not None:
            if kind_select is not None:
                msg = "`kind_select` must be `None` when `n_owners` is "
                msg += "specified. Got `{}`"
                raise ValueError(msg.format(kind_select))

            aq = aliased(selected, q.subquery())
            q = select(aq).join(Grant).where(Grant.id_user == self.id)

            cond_count = func.count(Grant.id_user)
            cond_having: ColumnElement[bool] = (
                (cond_count < n_owners)
                if not n_owners_levelsets
                else (cond_count == n_owners)
            )

            q = q.group_by(Document.id).having(cond_having)

        return q

    def q_select_documents_exclusive(
        self,
        document_uuids: Set[str] | None = None,
    ):
        return self.q_select_documents(
            level=Level.view,
            document_uuids=document_uuids,
            n_owners=1,
            n_owners_levelsets=True,
            exclude_deleted=True,
        )

    @classmethod
    def q_select_for_user(
        cls,
        user_uuid: str,
        uuids: Set[str] | None = None,
        *,
        exclude_deleted: bool = True,
    ) -> Select:
        # NOTE: This should get collaborators, etc. in the future.
        return cls.q_select_public(uuids, exclude_deleted)

    def q_conds_collections(
        self,
        uuid_collection: Set[str] | None = None,
        exclude_deleted: bool = True,
    ):
        conds = and_(Collection.id_user == User.id)
        if uuid_collection is not None:
            conds = and_(conds, Collection.uuid.in_(uuid_collection))
        if exclude_deleted:
            conds = and_(conds, Collection.deleted == false())
        return conds

    def q_collections(
        self,
        uuid_collection: Set[str] | None = None,
        exclude_deleted: bool = True,
    ) -> Select:
        conds = self.q_conds_collections(
            uuid_collection=uuid_collection,
            exclude_deleted=exclude_deleted,
        )
        return select(Collection).where(conds)

    # ----------------------------------------------------------------------- #
    # Chainables for endpoints.

    # NOTE: Chainable methods should be prefixed with `check_`.
    # def check_can_access_collection(self, collection: "Collection") -> Self:
    # return self

    def check_can_access_document(
        self,
        document: "Document",
        level: ResolvableLevel,
        *,
        grants: Dict[str, "Grant"] | None = None,
        grants_index: Literal["uuid_document", "uuid_user"] = "uuid_document",
        pending: bool = False,
        exclude_deleted: bool = True,
        validate: bool = True,
    ) -> Self:
        "Less horrific."

        def do_grant(grant):
            if grants is not None and grant is not None:
                grants[getattr(grant, grants_index)] = grant

        # NOTE: If the document is public and the level is view, then don't check.
        #       These lines can cause an issue where grants are not correctly
        #       added,
        # NOTE: Deleted, pending, insufficient should be included for better
        #       feedback. Do not exclude when adding.
        level, session = Level.resolve(level), self.get_session()
        q = self.q_select_grants(
            {document.uuid}, exclude_deleted=False, exclude_pending=False
        )

        res: Grant | None = session.execute(q).scalar()  # type: ignore
        detail: Dict[str, Any] = dict(
            uuid_user=self.uuid,
            uuid_document=document.uuid,
            level_grant_required=level,
        )

        if not validate:
            do_grant(res)
            return self

        match res:
            case None:
                raise ErrAccessDocumentGrantBase.httpexception(
                    "_msg_dne",
                    403,
                    level=level.name,
                    **detail,
                )
            case Grant(
                deleted=False, pending=True, pending_from=pending_from
            ) as grant if not pending:
                status, msg = 403, "_msg_grant_pending"
                if pending_from == PendingFrom.created:
                    status, msg = 500, "_msg_grant_pending_created"

                raise ErrAccessDocumentPending.httpexception(
                    msg,
                    status,
                    uuid_grant=grant.uuid,
                    level_grant=grant.level,
                    pending_from=pending_from,
                    **detail,
                )
            case Grant(deleted=deleted) as grant:
                if not deleted:
                    if grant.level.value < level.value:
                        raise ErrAccessDocumentGrantInsufficient.httpexception(
                            "_msg_insufficient",
                            403,
                            uuid_grant=grant.uuid,
                            level_grant=grant.level,
                            **detail,
                        )
                    do_grant(grant)
                    return self
                elif not exclude_deleted:
                    do_grant(grant)
                    return self
                else:
                    raise ErrAccessDocumentGrantBase.httpexception(
                        "_msg_dne",
                        410,
                        **detail,
                    )
            # NOTE: Server has become a teapot: Should never happen!
            case _:
                raise ErrAccessDocumentGrantBase.httpexception(
                    "_msg_inconcievable",
                    418,
                    **detail,
                    level_grant_required=Level.own,
                )

    def check_can_access_event(
        self,
        event: Event,
        status_code: int = 403,
    ) -> Self:
        if self.uuid != event.uuid_user:
            raise ErrAccessEvent.httpexception(
                "_msg_not_owner",
                403,
                uuid_event=event.uuid,
                uuid_user_token=self.uuid,
            )

        return self


class Collection(SearchableTableMixins, Base):
    __tablename__ = "collections"
    __kind__ = KindObject.collection

    id_user: Mapped[int] = mapped_column(
        ForeignKey("users.id", ondelete="CASCADE"),
        nullable=False,
    )
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(String(fields.LENGTH_NAME))
    description: Mapped[str] = mapped_column(
        String(fields.LENGTH_DESCRIPTION),
        nullable=True,
    )
    content: Mapped[Dict[str, Any]] = mapped_column(
        mysql.JSON, default=None, nullable=True
    )

    user: Mapped[User] = relationship(
        primaryjoin="User.id==Collection.id_user",
        back_populates="collections",
    )

    # NOTE: For all purposes, this is the parent side of the relationship. Do
    #       not cascade deletion of collections to documents.
    documents: Mapped[List["Document"]] = relationship(
        secondary=AssocCollectionDocument.__table__,
        back_populates="collections",
        # cascade="all, delete",
    )

    @property
    def uuid_user(self) -> str:
        session = self.get_session()
        q = select(User.uuid).where(User.id == self.id_user)
        res = session.execute(q).scalar()
        if res is None:
            raise ValueError("Inconcievable!")
        return res

    def q_conds_assignment(
        self,
        document_uuids: Set[str] | None = None,
        exclude_deleted: bool = True,
    ) -> ColumnElement[bool]:
        # NOTE: To add the conditions for document select (like level) use
        #       `q_conds_assoc`.
        cond = and_(AssocCollectionDocument.id_collection == self.id)
        if exclude_deleted:
            cond = and_(cond, AssocCollectionDocument.deleted == false())
        if document_uuids is not None:
            document_ids = Document.q_select_ids(document_uuids)
            cond = and_(cond, AssocCollectionDocument.id_document.in_(document_ids))
        # cond = and_(cond, self.q_conds(document_uuids, exclude_deleted))

        return cond

    def q_select_assignment(
        self,
        document_uuids: Set[str] | None = None,
        exclude_deleted: bool = True,
    ) -> Select:
        q = (
            select(
                AssocCollectionDocument,
                Document.uuid.label("uuid_document"),
                Collection.uuid.label("uuid_collection"),
            )
            .join(Collection)
            .join(Document)
        )
        q = q.where(self.q_conds_assignment(document_uuids, exclude_deleted))
        return q

    def q_select_documents(
        self,
        document_uuids: Set[str] | None = None,
        exclude_deleted: bool = True,
    ) -> Select:
        q = (
            select(Document)
            .join(AssocCollectionDocument)
            .where(self.q_conds_assignment(document_uuids, exclude_deleted))
        )
        return q

    @classmethod
    def q_select_for_user(
        cls,
        user_uuid,
        uuids: Set[str] | None = None,
        *,
        exclude_deleted: bool = True,
    ) -> Select:
        return (
            select(Collection)
            .join(User)
            .where(User.uuid == user_uuid)
            .where(cls.q_conds(uuids, exclude_deleted))
        )


class Document(SearchableTableMixins, Base):
    __tablename__ = "documents"
    __kind__ = KindObject.document

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(String(fields.LENGTH_NAME))
    description: Mapped[str] = mapped_column(
        String(fields.LENGTH_DESCRIPTION),
        nullable=True,
    )
    content: Mapped[Dict[str, Any]] = mapped_column(
        mysql.JSON, default=None, nullable=True
    )

    users: Mapped[List[User]] = relationship(
        secondary=AssocUserDocument.__table__,
        back_populates="documents",
        passive_deletes=True,
    )

    # NOTE: Child side of the relationship.
    collections: Mapped[List[Collection]] = relationship(
        secondary=AssocCollectionDocument.__table__,
        back_populates="documents",
        passive_deletes=True,
    )

    # ----------------------------------------------------------------------- #
    # Queries

    def q_conds_grants(
        self,
        user_uuids: Set[str] | None = None,
        level: ResolvableLevel | None = None,
        exclude_deleted: bool = True,
        pending: bool | None = None,
        exclude_pending: bool = True,
        pending_from: PendingFrom | None = None,
    ) -> ColumnElement[bool]:
        """

        :param pending: Specify a value for pending.
        :param exclude_pending: Specify if all grants should be returned
            regardless of their pending status.
        """
        cond = AssocUserDocument.id_document == self.id
        if exclude_deleted:
            cond = and_(
                cond,
                AssocUserDocument.deleted == false(),
                User.deleted == false(),
            )
        if user_uuids is not None:
            cond = and_(
                cond,
                AssocUserDocument.id_user.in_(
                    select(User.id).where(User.uuid.in_(user_uuids))
                ),
            )
        if level is not None:
            level = Level.resolve(level)
            cond = and_(cond, AssocUserDocument.level >= level.value)

        match (pending, exclude_pending):
            case (True, True):
                msg = "`pending` and `exclude_pending` cannot both be ``True``."
                raise ValueError(msg)
            case (bool() as pending, False):
                cond = and_(cond, Grant.pending == (true() if pending else false()))
            case (None | False, True):
                cond = and_(cond, Grant.pending == false())

        if pending_from is not None:
            cond = and_(cond, Grant.pending_from == pending_from)

        return cond

    def q_select_grants(
        self,
        user_uuids: Set[str] | None = None,
        level: ResolvableLevel | None = None,
        exclude_deleted: bool = True,
        exclude_pending: bool = True,
        pending: bool = False,
        pending_from: PendingFrom | None = None,
    ) -> Select:
        """Query to find grants (AssocUserDocument) for this document.

        To find grants for a user, see the same named method on :class:`User`.

        :param user_uuids: The uuids of the users to select for.
        :param level: The minimal level to select joined entries from.
        """
        conds = self.q_conds_grants(
            user_uuids=user_uuids,
            level=level,
            exclude_deleted=exclude_deleted,
            pending=pending,
            exclude_pending=exclude_pending,
            pending_from=pending_from,
        )
        return select(Grant).join(Document).join(User).where(conds)

    def q_select_users(
        self,
        user_uuids: Set[str] | None = None,
        level: ResolvableLevel | None = None,
        exclude_deleted: bool = True,
        exclude_pending: bool = True,
        pending: bool = False,
        pending_from: PendingFrom | None = None,
    ) -> Select:
        """Select user uuids for this document.

        For parameter descriptions, see :meth:`q_select_grants`.
        """
        q = (
            select(User)
            .join(AssocUserDocument)
            .where(
                self.q_conds_grants(
                    user_uuids=user_uuids,
                    level=level,
                    exclude_deleted=exclude_deleted,
                    exclude_pending=exclude_pending,
                    pending=pending,
                    pending_from=pending_from,
                )
            )
        )
        return q

    def q_conds_assignment(
        self,
        collection_uuids: Set[str] | None = None,
        exclude_deleted: bool = True,
    ) -> ColumnElement[bool]:
        # NOTE: To add the conditions for document select (like level) use
        #       `q_conds_assoc`.
        cond = and_(AssocCollectionDocument.id_document == self.id)
        if exclude_deleted:
            cond = and_(
                cond,
                AssocCollectionDocument.deleted == false(),
            )
        if collection_uuids is not None:
            collection_ids = Collection.q_select_ids(collection_uuids)
            cond = and_(cond, AssocCollectionDocument.id_collection.in_(collection_ids))

        return cond

    def q_select_assignment(
        self,
        collection_uuids: Set[str] | None = None,
        exclude_deleted: bool = True,
    ) -> Select:
        q = (
            select(
                AssocCollectionDocument,
                Document.uuid.label("uuid_document"),
                Collection.uuid.label("uuid_collection"),
            )
            .join(Document)
            .join(Collection)
        )
        q = q.where(self.q_conds_assignment(collection_uuids, exclude_deleted))
        return q

    def q_select_collections(
        self,
        collection_uuids: Set[str] | None = None,
        exclude_deleted: bool = True,
    ) -> Select:
        q = (
            select(Collection)
            .join(Assignment)
            .where(self.q_conds_assignment(collection_uuids, exclude_deleted))
        )
        return q

    @classmethod
    def q_select_documents(
        cls,
        user: User,
        uuid_documents: Set[str] | None = None,
        exclude_deleted: bool = True,
        before: int | None = None,
        after: int | None = None,
        limit: int | None = None,
    ) -> Select:
        conds = []
        if before is not None:
            conds.append(Event.timestamp <= before)
        if after is not None:
            conds.append(Event.timestamp >= after)

        q_uuids = (
            select(Event.uuid_obj)
            .where(
                Event.kind_obj == KindObject.document,
                Event.kind == KindEvent.update,
                Event.detail == Event.detail_document_editted,
                *conds,
            )
            .distinct()
        )
        if limit is not None:
            q_uuids = q_uuids.limit(limit)

        q = select(cls).where(cls.uuid.in_(q_uuids))
        if exclude_deleted:
            q = q.where(cls.deleted == false())
        if uuid_documents:
            q = q.where(cls.uuid.in_(uuid_documents))

        conds = user.q_conds_grants(uuid_documents)
        q = q.join(Grant).join(User).where(*conds)

        return q

    @classmethod
    def q_select_for_user(
        cls,
        user_uuid: str,
        uuids: Set[str] | None = None,
        *,
        exclude_deleted: bool = True,
    ) -> Select:
        return (
            select(Document)
            .join(AssocUserDocument)
            .join(User)
            .where(User.uuid == user_uuid)
            .where(cls.q_conds(uuids, exclude_deleted))
        )


# --------------------------------------------------------------------------- #
# After the matter constants and types.


class Tables(enum.Enum):
    assignments = AssocCollectionDocument
    grants = AssocUserDocument
    events = Event
    users = User
    collections = Collection
    documents = Document


Assignment = AssocCollectionDocument
Grant = AssocUserDocument
AnyModelBesidesEvent = (
    AssocUserDocument | AssocCollectionDocument | User | Collection | Document
)
AnyModel = Event | AnyModelBesidesEvent
AnyModelType = Type[AnyModel]
ResolvableModel = AnyModelType | KindObject


def resolve_model(
    resolvable: ResolvableModel | str, exclude_secondary: bool = False
) -> AnyModelType:
    match resolvable:
        case KindObject() as kind:
            Model = Tables[Plural[kind.name].value].value
        case type(__tablename__=str()) as Model:
            ...
        case str() as v:
            kind = KindObject(v)
            return resolve_model(kind)
        case bad:
            raise ValueError(f"Cannot resolve `{bad}` to model.")

    if exclude_secondary and Model in (Grant, Assignment):
        raise ValueError(f"Excluding secondary model `{Model.__tablename__}`.")

    return Model


# Resolvables

T_Resolvable = TypeVar(
    "T_Resolvable",
    AssocUserDocument,
    AssocCollectionDocument,
    User,
    Collection,
    Document,
    Event,
)

ResolvableSingular: TypeAlias = T_Resolvable | str
ResolvableMultiple: TypeAlias = Tuple[T_Resolvable, ...] | Set[str]
Resolvable: TypeAlias = T_Resolvable | str | Tuple[T_Resolvable, ...] | Set[str]

ResolvableSourceAssignment: TypeAlias = (
    ResolvableSingular[Collection] | ResolvableSingular[Document]
)
ResolvableTargetAssignment: TypeAlias = (
    ResolvableMultiple[Document] | ResolvableMultiple[Collection]
)
ResolvableSourceGrant: TypeAlias = (
    ResolvableSingular[User] | ResolvableSingular[Collection]
)
ResolvableTargetGrant: TypeAlias = (
    ResolvableMultiple[User] | ResolvableMultiple[Collection]
)


# Resolved result.
ResolvedRaw = Tuple[T_Resolvable, ...]  # | Resolvable[T_Resolvable]
ResolvedRawUser = ResolvedRaw[User]
ResolvedRawCollection = ResolvedRaw[Collection]
ResolvedRawDocument = ResolvedRaw[Document]

ResolvedRawAssignmentDocument = Tuple[Document, Tuple[Collection, ...]]
ResolvedRawAssignmentCollection = Tuple[Collection, Tuple[Document, ...]]
ResolvedRawAssignment = ResolvedRawAssignmentDocument | ResolvedRawAssignmentCollection

ResolvedRawGrantUser = Tuple[User, Tuple[Document, ...]]
ResolvedRawGrantDocument = Tuple[Document, Tuple[User, ...]]
ResolvedRawGrant = ResolvedRawGrantUser | ResolvedRawGrantDocument

ResolvedRawAny = (
    ResolvedRawCollection
    | ResolvedRawUser
    | ResolvedRawDocument
    | ResolvedRawAssignment
    | ResolvedRawGrant
)


def uuids(vs: Dict[str, T_Resolvable] | Resolvable[T_Resolvable]) -> Set[str]:
    match vs:
        case tuple():
            return {vv.uuid for vv in vs}
        case set():
            return vs
        case str():
            return {vs}
        case dict():
            return set(item.uuid for item in vs.values())
        case _ as item:
            return {item.uuid}


__all__ = (
    "Base",
    "User",
    "Collection",
    "AssocCollectionDocument",
    "AssocUserDocument",
    "Document",
    "Tables",
    "KindEvent",
    "KindObject",
    "Level",
    "LevelStr",
    "KindObject",
    "ChildrenUser",
    "ChildrenCollection",
    "ChildrenDocument",
)



# =========================================================================== #
import json
import traceback
from typing import Annotated, Generator, List, Set

from fastapi import FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse
from fastapi.routing import APIRoute
from fastapi.staticfiles import StaticFiles
from pydantic import BaseModel, ConfigDict, Field, TypeAdapter
from starlette.middleware import Middleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.routing import Mount

# --------------------------------------------------------------------------- #
from captura import __version__, util
from captura.controllers.access import H
from captura.depends import DependsAccess

from .assignments import CollectionAssignmentView, DocumentAssignmentView
from .auth import AuthViewAuth0, AuthViewPytest
from .base import BaseView, OpenApiTagMetadata, OpenApiTags
from .collections import CollectionView
from .documents import DocumentView
from .events import EventView
from .grants import DocumentGrantView, UserGrantView
from .users import UserView

logger = util.get_logger(__name__)

description: str = """
An API for storing, rendering, editting, and sharing text documents and other 
deffered objects. 


## Sharing

Documents can be public or private and owners can grant to other users
the ability to view, modify, or own the document.


## Deffered Objects

Soon the Captura API will support deffering objects to other APIs. More
or less this model will be 'sharing as a service'.
"""


class AppRouteInfo(BaseModel):
    model_config = ConfigDict(from_attributes=True)
    path: Annotated[str, Field()]
    name: Annotated[str, Field()]
    methods: Annotated[Set[H], Field()]


class AppView(BaseView):
    # NOTE: The following are helpful, I don't want to do this right now but
    #       it will be useful for later.
    #
    #       - https://swagger.io/docs/open-source-tools/swagger-ui/usage/configuration/
    #       - https://fastapi.tiangolo.com/how-to/configure-swagger-ui/?h=swagger
    #
    # TODO: Make swagger docs better. Add better examples for requests and
    #       responses, etc.
    #
    #       - https://fastapi.tiangolo.com/reference/openapi/models/
    #
    # NOTE: About static files:
    #
    #       - https://fastapi.tiangolo.com/tutorial/static-files/
    #
    #       I'm aware the using `routes` is a bit taboo, but I think this is a
    #       cleaner pattern.
    routes = []
    if util.path.exists(util.PATH_STATIC):
        view_static = StaticFiles(directory=util.PATH_STATIC)
        routes.append(Mount("/static", view_static))

    view_router = FastAPI(
        title="Captura Document Automation, Sharing, and Organization API.",
        description=description,
        version=__version__,
        contact=dict(
            name="Adrian Cederberg",
            url="https://github.com/acederberg",
            email="adrn.cederberg123@gmail.com",
        ),
        openapi_tags=OpenApiTagMetadata,
        swagger_ui_parameters={"syntaxHighlight.theme": "obsidian"},
        routes=routes,  # type: ignore
        middleware=[
            Middleware(
                SessionMiddleware,
                secret_key=util.SESSION_SECRET,
            )
        ],
    )  # type: ignore

    view_routes = {
        "get_index": {
            "url": "/",
            "tags": [OpenApiTags.html],
            "name": "Index Page",
            "description": "For humans.",
        },
        "get_routes": {
            "url": "/routes",
            "tags": [OpenApiTags.admin],
            "name": "Routes Directory",
            "description": "For text based clients that won't like using help.",
        },
    }
    view_children = {
        # "": EventSearchView,
        "/grants/documents": DocumentGrantView,
        "/grants/users": UserGrantView,
        "/assignments/documents": DocumentAssignmentView,
        "/assignments/collections": CollectionAssignmentView,
        "/users": UserView,
        "/collections": CollectionView,
        "/documents": DocumentView,
        "/auth": AuthViewPytest,
        "/events": EventView,
        "": AuthViewAuth0,
    }

    @classmethod
    def get_index(cls, request: Request):
        return cls.view_templates.TemplateResponse(request, "index.j2", {})

    @classmethod
    def get_routes(
        cls,
        access: DependsAccess,
        methods: Set[H] | None = None,
        names: Set[str] | None = None,
        names_fragment: str | None = None,
        paths_fragment: str | None = None,
    ) -> List[AppRouteInfo]:
        if not access.token:
            raise HTTPException(403, detail="Token required.")

        # NOTE: Type ignored for metaclass considerations.
        items: Generator[APIRoute, None, None]
        items = (
            item
            for item in cls.view_router.routes  # type: ignore
            if isinstance(item, APIRoute)
        )

        if methods is not None:
            items = (
                item
                for item in items
                if all(method in methods for method in item.method)  # type: ignore
                # hasattr(item, "methods")
                # and
            )

        match (names, names_fragment):
            case (set(), None):
                items = (item for item in items if item.name in names)  # type: ignore
            case (None, str()):
                items = (item for item in items if names_fragment in item.name)
            case (None, None):
                pass
            case _:
                msg = "Only one of `names` and `names_fragment` may be "
                msg += "specified."
                raise HTTPException(422, detail=dict(msg=msg))

        if paths_fragment is not None:
            items = (item for item in items if paths_fragment in item.path)

        t = TypeAdapter(List[AppRouteInfo])
        return t.validate_python(items)


# =========================================================================== #


# TODO: The traceback should not show up in prod unless the status `500`.
#       In fact, the traceback should be stored by the logger instead.
# TODO: Figure out how to use config to turn this on or off. Unfortunately
#       exception handlers do not use dependencies and it is hard to build
#       dependecies from the request directly.
@AppView.view_router.exception_handler(HTTPException)  # type: ignore
async def http_exception_handler(request: Request, exc: HTTPException):
    # https://github.com/encode/uvicorn/blob/master/uvicorn/protocols/http/h11_impl.py
    if util.VERBOSE_HTTPEXCEPTIONS:
        traceback.print_exc()

    if util.VERBOSE:
        if request.client:
            host, port = request.client.host, request.client.port
            logger.info("%s:%s - %s", host, port, json.dumps(exc.detail, indent=2))
        else:
            logger.debug("No client for request.")

    return JSONResponse(
        exc.detail,
        exc.status_code,
    )


__all__ = (
    "DocumentGrantView",
    "UserGrantView",
    "CollectionView",
    "EventView",
    "DocumentView",
    "UserView",
    "BaseView",
)



# =========================================================================== #
import json
import logging
from typing import List, Set

LOG_RECORD_BUILTIN_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}


# Thanks mCoding: https://www.youtube.com/watch?v=9L77QExPmI0
class JSONFormatter(logging.Formatter):
    fmt_keys: Set[str]

    @property
    def fmt_keys_default(self) -> Set[str]:
        return {
            "levelname",
            "timestamp",
            "name",
            "module",
            "funcName",
            "lineno",
            "threadName",
        }

    def __init__(self, *, fmt_keys: List[str] | None = None):
        super().__init__()
        self.fmt_keys = (
            (set(fmt_keys) & LOG_RECORD_BUILTIN_ATTRS)
            if fmt_keys is not None
            else self.fmt_keys_default
        )
        if "message" in self.fmt_keys:
            raise ValueError("Cannot specify `message` in format keys.")

    def format(self, record: logging.LogRecord) -> str:
        line = {key: getattr(record, key, None) for key in self.fmt_keys}
        line.update(msg=record.getMessage())
        if record.exc_info is not None:
            line.update(exc_info=self.formatException(record.exc_info))
        if record.stack_info is not None:
            line.update(stack_info=self.formatStack(record.stack_info))

        return json.dumps(line, default=str)



# =========================================================================== #
import secrets
from http import HTTPMethod
from random import choice, randint
from typing import (
    Annotated,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    NotRequired,
    Self,
    Set,
    Tuple,
    Type,
    TypedDict,
)

import httpx
import yaml
from sqlalchemy import (
    delete,
    desc,
    false,
    func,
    literal_column,
    or_,
    select,
    true,
    update,
)

# from sqlalchemy.orm import session as sqa_session
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker as sqa_sessionmaker
from typing_extensions import Doc

# --------------------------------------------------------------------------- #
from captura import util
from captura.auth import Auth, Token, TokenPermissionTier
from captura.controllers.access import Access
from captura.controllers.base import (
    BaseResolved,
    BaseResolvedPrimary,
    BaseResolvedSecondary,
    Data,
    ResolvedDocument,
    ResolvedUser,
    T_ResolvedPrimary,
)
from captura.controllers.create import Create
from captura.controllers.delete import Delete
from captura.fields import KindEvent, KindObject, Level
from captura.models import (
    Assignment,
    AssocCollectionDocument,
    AssocUserDocument,
    Base,
    Collection,
    Document,
    Event,
    Grant,
    PendingFrom,
    ResolvableModel,
    ResolvedRawCollection,
    ResolvedRawDocument,
    User,
    resolve_model,
    uuids,
)
from captura.schemas import mwargs
from legere import ConsoleHandler, ContextData, Requests
from legere.config import Config as ClientConfig
from legere.config import ProfileConfig, UseConfig
from simulatus.config import ConfigSimulatus, DummyConfig
from simulatus.mk import Mk, combos
from simulatus.reports import ReportController

util.setup_logging(util.Path.base("logging.test.yaml"))
logger = util.get_logger(__name__)

# =========================================================================== #
# Providers


class GetPrimaryKwargs(TypedDict):
    public: NotRequired[bool | None]
    deleted: NotRequired[bool | None]
    uuids: NotRequired[Set[str] | None]
    retry: NotRequired[bool]
    retry_count_max: NotRequired[int]
    allow_empty: NotRequired[bool]


class GetPrimaryDataKwargs(TypedDict):
    n: NotRequired[int | None]
    get_primary_kwargs: NotRequired[GetPrimaryKwargs]
    other: NotRequired[bool]
    callback: NotRequired[Callable[[Dict[str, Any], Tuple[Any, ...]], None]]


class BaseDummyProvider:
    dummy_kinds: ClassVar[Set[KindObject]] = {
        KindObject.user,
        KindObject.collection,
        KindObject.document,
        KindObject.grant,
        KindObject.assignment,
        KindObject.event,
    }

    client_config_cls: Type
    config: ConfigSimulatus
    dummy: DummyConfig
    session: Session
    user: User
    auth: Auth

    def __init__(
        self,
        config: ConfigSimulatus,
        session: Session,
        *,
        user: User,
        auth: Auth | None = None,
        client_config_cls: Type | None = None,
    ):
        self.client_config_cls = client_config_cls or ClientConfig
        self.config = config
        self.dummy = config.dummy
        self.auth = auth if auth is not None else Auth.forPyTest(config)
        self.session = session
        self.user = user

    # ----------------------------------------------------------------------- #
    # Functions derivative of ``Mk``.

    def mk_user(self) -> User:
        user = Mk.user()
        user.admin = False
        user.deleted = False
        return user

    def mk_documents(self) -> Tuple[Document, ...]:
        # NOTE: Grants cereated in :func:`mk_grants`.
        logger.debug("Making documents.")

        assert isinstance(self.dummy, DummyConfig)
        a, b = self.dummy.documents.minimum, self.dummy.documents.maximum
        return tuple(Mk.document() for _ in range(randint(a, b)))

    def mk_assignments(
        self,
        documents: Tuple[Document, ...],
        collections: Tuple[Collection, ...],
    ) -> Tuple[Assignment, ...]:
        # NOTE: Iter over collections first so that every document has a chance
        #       belong to each collection.
        # TODO: Make it such that this is not restricted to a user only having
        #       its collections made exclusively of its own documents.
        logger.debug("Making assignments...")
        assignments: Tuple[Assignment, ...] = tuple(
            Assignment(
                id_document=document.id,
                id_collection=collection.id,
                deleted=False,
            )
            for collection in collections
            for document in documents
            if randint(0, 2)
        )
        self.session.add_all(assignments)
        return assignments

    def mk_grants(
        self,
        documents: Tuple[Document, ...],
        exclude_grants_self: bool = False,
        exclude_grants_create: bool = False,
    ):
        dummies = self.dummy

        # NOTE: Adding grants to the database can be a real pain in the ass
        #       because the primary key of the grants table is infact the
        #       ``uuid`` column and not the associated tables since the table
        #       is a tree. For this reason it is advised that a delete
        #       statement be emitted in addition to these queries.
        #
        # TLDR: Do not use ``session.merge``, just use ``DELETE`` or use
        #       options to avoid causing duplicate ``(id_user, id_document)``
        #       keys.
        if not exclude_grants_create:
            logger.debug("Making grants for documents created by self.")
            grants_created = tuple(
                Grant(
                    level=Level.own,
                    id_user=self.user.id,
                    id_document=document.id,
                    deleted=False,
                    pending=False,
                    pending_from=PendingFrom.created,
                )
                for document in documents
            )
        else:
            grants_created = tuple()

        # NOTE: Add grants for documents of other users.
        # NOTE: Grants giving access to documents that exist already.
        if not exclude_grants_self:
            logger.debug("Making grants for `%s` on documents created by others.")
            a, b = dummies.grants.minimum_self, dummies.grants.maximum_self
            q_ids_has_grants = (
                select(Document.id)
                .join(Grant)
                .group_by(Grant.id_document)
                .having(func.count(Grant.id_user) > 0)
                .where(Grant.id_user == self.user.id)
            )
            q_grants_share = (
                select(Grant)
                .join(Document)
                .where(
                    Document.id.not_in(q_ids_has_grants),
                    Grant.pending_from == PendingFrom.created,
                )
                .limit(b)
                .order_by(func.random())
            )
            grants_share = tuple(self.session.scalars(q_grants_share))

            if (n := len(grants_share)) < a:
                logger.warning(
                    "Only `%s` documents for user `%s`.",
                    n,
                    self.user.uuid,
                )

            grants_self = tuple(
                Grant(
                    uuid=secrets.token_urlsafe(8),
                    id_user=self.user.id,
                    id_document=grant.id_document,
                    uuid_parent=grant.uuid,
                    **kwargs,
                )
                for (grant, kwargs) in zip(grants_share, combos())
            )

        else:
            grants_self = tuple()

        # NOTE: Grants to others on documents created. ``grants_source`` is
        #       included since there is the option to not make ``created`` grants
        #       for the provided documents.
        logger.debug("Making grants for others on documents created `%s`.")

        q_ids_users = (
            select(User.id)
            .where(User.id != self.user.id)
            .limit(
                randint(
                    dummies.grants.minimum_other,
                    dummies.grants.maximum_other,
                )
            )
        )
        id_users_other = set(self.session.scalars(q_ids_users))

        if not grants_created:
            grants_source = tuple(map(self.get_document_grant, documents))
            assert len(grants_source) == len(documents)
        else:
            grants_source = grants_created

        grants_other = tuple(
            Grant(
                uuid=secrets.token_urlsafe(8),
                level=choice(list(Level)),
                id_user=id_user,
                id_document=grant.id_document,
                uuid_parent=grant.uuid,
                deleted=bool(randint(0, 1)),
                pending=bool(randint(0, 1)),
                pending_from=choice((PendingFrom.grantee, PendingFrom.granter)),
            )
            for grant in grants_source
            for id_user in id_users_other
        )

        return grants_created + grants_self + grants_other

    def mk_collections(self) -> Tuple[Collection, ...]:
        logger.debug("Making collections...")
        dummies = self.dummy
        a, b = dummies.documents.minimum, dummies.documents.maximum
        return tuple(Mk.collection(user=self.user) for _ in range(randint(a, b)))

    def mk(
        self,
        documents: ResolvedRawDocument | None = None,
        collections: ResolvedRawCollection | None = None,
        # edits: ResolvedRawCollection | None = None,
        create_grants: None | Dict[str, Any] = dict(),  # options for mk_grants
        create_assignments: bool = True,
        create_documents_unique: bool = True,
    ) -> Dict[str, Set[str]]:
        """Set up a new user or add data."""

        # NOTE: Create dummy users, documents, and collections. Grants and
        #       assignments will be made subsequently.
        data: Dict[str, Set[str]] = dict()
        session = self.session
        session.add(self.user)
        session.commit()
        session.refresh(self.user)

        data.update(user={self.user.uuid})

        if documents is None:
            session.add_all(documents := self.mk_documents())
            session.commit()
            data.update(documents=uuids(documents))

        if collections is None:
            session.add_all(collections := self.mk_collections())
            session.commit()
            data.update(collections=uuids(collections))

        # NOTE: Session.merge only works for assignments and not grants because
        #       assignments uses the association primary keys to form a its
        #       primary key while grants uses the uuid. See :func:`mk_grants`.
        if create_assignments:
            assert documents is not None, "Documents required."
            assert collections is not None, "Collections required."
            tuple(
                map(
                    lambda item: session.merge(item),
                    (assignments := self.mk_assignments(documents, collections)),
                )
            )
            session.commit()
            data.update(assignments=uuids(assignments))

        if create_grants is not None:
            assert documents is not None, "Documents required."
            grants = self.mk_grants(documents, **create_grants)

            tuple(
                map(
                    lambda item: session.merge(item),
                    grants,
                )
            )
            session.commit()
            data.update(grants=uuids(grants))

        # NOTE: Create some uniquely owned documents (important for tests of
        #       deltion cascading configuration). It is important to note
        #       that these documents are only gaurenteed to be unique when the
        #       dummy is the most recently generated.
        if create_documents_unique:
            documents_uniq = self.mk_documents()
            session.add_all(documents_uniq)
            session.commit()

            data["documents"] |= uuids(documents_uniq)

        return data

    # ----------------------------------------------------------------------- #
    # Getters

    def randomize_primary(
        self,
        model: ResolvableModel,
        uuids: Set[str] | None = None,
    ) -> None:
        Model = resolve_model(model)

        q = update(Model).where(func.random() < 0.5)
        if uuids is not None:
            q = q.where(Model.uuid.in_(uuids))
        tuple(
            self.session.execute(q.values(deleted=deleted, public=public))
            for deleted in (0, 1)
            for public in (0, 1)
        )
        self.session.commit()

    def randomize_grants(self):
        logger.debug("Randomizing grants for `%s`.", self.user.uuid)
        q_uuids = select(Grant.uuid).where(
            Grant.id_user == self.user.id, Grant.pending_from != PendingFrom.created
        )
        uuids = self.session.scalars(q_uuids)

        for pending in (0, 1):
            for pending_from in (PendingFrom.grantee, PendingFrom.granter):
                q = (
                    update(Grant)
                    .values(pending=pending, pending_from=pending_from)
                    .where(Grant.uuid.in_(uuids), func.random() < 0.5)
                )
                self.session.execute(q)

        self.session.commit()

    def get_primary(
        self,
        model: ResolvableModel,
        n: int,
        *,
        public: bool | None = None,
        deleted: bool | None = False,
        uuids: Set[str] | None = None,
        callback: Callable[[Any], Any] | None = None,
        retry_callback: Callable[[], Tuple[T_ResolvedPrimary, ...]] | None = None,
        retry_count_max: int = 1,
        retry: bool = True,
        allow_empty: bool = False,
    ) -> Tuple[T_ResolvedPrimary, ...]:
        Model = resolve_model(model)

        if depr := tuple(map(lambda v: v is not None, (retry_count_max, retry))):
            logger.warning(f"Retry args `{depr}` are depricated.")

        logger.debug("Getting data of kind `%s`.", Model.__tablename__)
        conds = list()
        if public is not None and hasattr(Model, "public"):
            bool_ = true() if public else false()
            conds.append(Model.public == bool_)
        if deleted is not None:
            bool_ = true() if deleted else false()
            conds.append(Model.deleted == bool_)
        if uuids is not None:
            conds.append(Model.uuid.in_(uuids))

        q = select(Model).where(*conds).order_by(func.random()).limit(n)
        if callback:
            q = callback(q)

        # NOTE: Move assertions here to avoid preflight checks in tests.
        models = tuple(self.session.scalars(q))
        if not allow_empty and len(models) == 0:

            ReportController(self.session).create(
                note=f"Failed to find `{Model.__tablename__}`. ",
                user=self.user,
                fn="Dummy.get_primary",
                module="dummy.__init__",
                tags=["pytest", "get_primary", "fail"],
            )
            if retry_callback is not None:
                return retry_callback()

            raise AssertionError(
                f"Could not find test data for `{self.user.uuid}` table "
                f"`{Model.__tablename__}`. Offending query:\n"
                f"{util.sql_render(self.session, q)}"
            )

        if deleted is not None:
            logger.debug("Checking deleted.")
            assert all(mm.deleted is deleted for mm in models)

        if public is not None:
            logger.debug("Checking public.")
            assert all(getattr(mm, "public", None) is public for mm in models)
        return models  # type: ignore

    def get_users(
        self,
        n: int,
        get_primary_kwargs: GetPrimaryKwargs = GetPrimaryKwargs(),
        *,
        other: bool = False,
    ) -> Tuple[User, ...]:
        callback = None
        if other:
            callback = lambda q: q.where(User.id != self.user.id)  # noqa: E731

        return self.get_primary(
            User,
            n,
            callback=callback,
            **get_primary_kwargs,
        )

    def get_users_data(
        self,
        n: int | None = None,
        get_primary_kwargs: GetPrimaryKwargs = GetPrimaryKwargs(),
        *,
        other: bool = False,
    ) -> Data[ResolvedUser]:
        return self.get_data_primary(User, n, get_primary_kwargs, other=other)

    # ----------------------------------------------------------------------- #

    def get_documents(
        self,
        n: int,
        get_primary_kwargs: GetPrimaryKwargs = GetPrimaryKwargs(),
        *,
        level: Level | None = None,
        other: bool | None = False,
        **kwargs,
    ) -> Tuple[Document, ...]:
        """Gets user documents by default.

        When :param:`other` is ``None``, the ownership of documents should be
        treated arbitrarily.
        """

        if get_primary_kwargs.get("deleted"):
            kwargs["exclude_deleted"] = False

        def callback(q):
            if other is None:
                ...
            elif not other:
                assert level is not None, "Level is required when `other` is `False`."
                conds = self.user.q_conds_grants(level=level, **kwargs)
                q = q.join(Grant).where(*conds)
            else:
                q_has_grants = (
                    select(Grant.id_document)
                    .where(Grant.id_user == self.user.id)
                    .group_by(Grant.id_document)
                    .having(func.count(Grant.uuid) > 0)
                )
                q = q.where(Document.id.not_in(q_has_grants))

            return q

        def get_documents_retry_callback():
            if other:
                raise ValueError("Not yet supported.")

            pending_from = kwargs.get("pending_from")
            if pending_from == PendingFrom.created:
                assert n == 1
                uuid = secrets.token_urlsafe(8)
                document = Document(
                    uuid=uuid,
                    name=f"document-get-documents-retry-callback-{uuid}",
                    description="Made because of failure of `get_primary`.",
                )
                self.session.add(document)
                self.session.commit()
                self.session.refresh(document)

                self.session.add(
                    Grant(
                        uuid=secrets.token_urlsafe(8),
                        id_document=document.id,
                        id_user=self.user.id,
                        level=level or Level.own,
                        pending_from=PendingFrom.created,
                        pending=kwargs.get("pending", False),
                        deleted=kwargs.get("deleted", False),
                    )
                )
                self.session.commit()  # NOTE: It might look stupid to do this but it works.
                return (document,)

            documents = self.get_documents(n, other=True)
            grants = tuple(
                Grant(
                    uuid=secrets.token_urlsafe(8),
                    id_document=document.id,
                    id_user=self.user.id,
                    level=level or Level.own,
                    pending_from=PendingFrom.created,
                    pending=kwargs.get("pending", False),
                    deleted=kwargs.get("deleted", False),
                )
                for document in documents
            )
            self.session.add_all(grants)
            self.session.commit()

            return documents

        return self.get_primary(
            Document,
            n,
            callback=callback,
            retry_callback=get_documents_retry_callback,
            **get_primary_kwargs,
        )

    def get_documents_data_callback(
        self,
        data_raw: Dict[str, Any],
        data_items: Tuple[Document, ...],
    ) -> None:
        uuid_documents = uuids(data_items)
        q_grants = (
            select(Grant)
            .join(Document)
            .where(Document.uuid.in_(uuid_documents), Grant.id_user == self.user.id)
        )
        data_raw["grants"] = {
            gg.uuid_document: gg for gg in self.session.scalars(q_grants)
        }
        data_raw["token_user_grants"] = data_raw["grants"]
        data_raw["uuids"] = uuid_documents

    def get_documents_data(
        self,
        n: int | None = None,
        get_primary_kwargs: GetPrimaryKwargs = GetPrimaryKwargs(),
        *,
        other: bool = False,
        **kwargs,
    ) -> Data[ResolvedDocument]:
        return self.get_data_primary(
            Document,
            n,
            get_primary_kwargs,
            other=other,
            callback=self.get_documents_data_callback,
            **kwargs,
        )

    # ----------------------------------------------------------------------- #

    def get_collections_retry_callback(self):
        # NOTE: Far less stupid.
        session = self.session
        logger.warning("Calling `get_collections_retry_callback`.")

        collection = Mk.collection(id_user=self.user.id)
        session.add(collection)
        session.commit()
        session.refresh(collection)

        # NOTE: Because ids must exist (for now, until only uuids).
        documents = self.get_documents(n=5, other=True)
        assignments = self.mk_assignments(
            documents=documents, collections=(collection,)
        )
        session.add_all(assignments)
        session.commit()

        return (collection,)

    def get_collections(
        self,
        n: int,
        get_primary_kwargs: GetPrimaryKwargs = GetPrimaryKwargs(),
        other: bool | None = False,
        order_by_document_count: bool = True,
    ) -> Tuple[Collection, ...]:
        def callback(q):
            match other:
                case True:
                    cond_other = Collection.id_user != self.user.id
                case False:
                    cond_other = Collection.id_user == self.user.id
                case _:
                    cond_other = true()

            if order_by_document_count:
                q_ids = (
                    select(Collection.id.label("id_collection"))
                    .join(Assignment)
                    .group_by(Collection.id)
                    .having(func.count(Assignment.id_document) > 0)
                    .where(cond_other)
                    .order_by(desc(func.count(Assignment.id_document)))
                )
                q = q.where(Collection.id.in_(q_ids))

            return q

        return self.get_primary(
            Collection,
            n,
            callback=callback,
            retry_callback=self.get_collections_retry_callback,
            **get_primary_kwargs,
        )

    def get_collections_data(
        self,
        n: int | None = None,
        get_primary_kwargs: GetPrimaryKwargs = GetPrimaryKwargs(),
        *,
        other: bool = False,
    ) -> Data[ResolvedUser]:
        return self.get_data_primary(Collection, n, get_primary_kwargs, other=other)

    # ----------------------------------------------------------------------- #

    def get_document_grant(self, document: Document) -> Grant:
        grant = self.session.scalar(
            select(Grant).where(
                Grant.id_document == document.id, Grant.id_user == self.user.id
            )
        )
        if grant is None:
            raise AssertionError("Grant should have been created for document.")
        return grant

    def get_events(
        self,
        n: int,
        get_primary_kwargs: GetPrimaryKwargs = GetPrimaryKwargs(),
        other: bool | None = False,
        uuid_obj: str | None = None,
        kind_obj: KindObject | None = None,
        kind: KindEvent | None = None,
    ) -> Tuple[Event, ...]:
        """It is possible that a user has no events. To generate events,
        do something with the dummy first.
        """

        def callback(q):
            if other is None:
                ...
            elif not other:
                q = q.where(Event.uuid_user == self.user.uuid)
            else:
                q = q.where(Event.uuid_user != self.user.uuid)

            if kind_obj is not None:
                q = q.where(Event.kind_obj == kind_obj)
            if kind is not None:
                q = q.where(Event.kind == kind)
            if uuid_obj is not None:
                q = q.where(Event.uuid_obj == uuid_obj)

            return q

        if "retry" not in get_primary_kwargs:
            get_primary_kwargs["retry"] = False
        if "allow_empty" not in get_primary_kwargs:
            get_primary_kwargs["allow_empty"] = True

        return self.get_primary(
            Event,
            n,
            callback=callback,
            **get_primary_kwargs,
        )  # type: ignore

    def get(
        self,
        Model: ResolvableModel,
        n: int,
        get_primary_kwargs: GetPrimaryKwargs = GetPrimaryKwargs(),
        *,
        other: bool = False,
        **kwargs,
    ) -> Tuple[T_ResolvedPrimary, ...]:
        """This exists because :meth:`get_primary` will not include all of the nice
        callbacks and keywords.

        Intended for generic cases, like :meth:`data_primary`.
        """
        meth: Callable
        match (resolve_model(Model)).__kind__:
            case KindObject.user:
                meth = self.get_users
            case KindObject.document:
                meth = self.get_documents
            case KindObject.collection:
                meth = self.get_collections
            case KindObject.event:
                meth = self.get_events
            case bad:
                raise ValueError(f"Cannot construct data of kind `{bad.name}`.")

        return meth(n, get_primary_kwargs, other=other, **kwargs)  # type: ignore

    def get_data_primary(
        self,
        Model: ResolvableModel,
        n: int | None = None,
        get_primary_kwargs: GetPrimaryKwargs = GetPrimaryKwargs(),
        *,
        other: bool = False,
        callback: Callable[[Dict[str, Any], Tuple[Any, ...]], None] | None = None,
        **kwargs,
    ):
        model = resolve_model(Model)
        data_primary = self.get(
            model,
            n if n is not None else randint(5, 15),
            get_primary_kwargs,
            other=other,
            **kwargs,
        )

        ResolvedModel = BaseResolvedPrimary.get(model.__kind__)
        data = {ResolvedModel._attr_name_targets: data_primary}
        if callback is not None:
            callback(data, data_primary)

        return mwargs(
            Data,
            data=ResolvedModel.model_validate(data),
        )

    def get_data_grant_document(
        self,
        # get_kwargs_document: Dict[str, Any] = dict(level=Level.own),
        get_kwargs_users: Dict[str, Any] = dict(),
        n: int | None = None,
    ):
        # NOTE: Document must be owned by :attr:`user`.
        def callback(raw: Dict[str, Any]) -> None:
            raw["token_user_grants"] = {
                self.user.uuid: self.get_document_grant(raw["document"]),
            }

        return self.get_data_secondary(
            Document,
            User,
            # get_kwargs_target=get_kwargs_document,
            source=self.get_documents(1, level=Level.own)[0],
            get_kwargs_target=get_kwargs_users,
            callback=callback,
            n=n,
        )

    def get_data_grant_user(
        self,
        get_kwargs_document: Dict[str, Any] = dict(other=None),
        n: int | None = None,
    ):
        # NOTE: For now, ``token_user_grants`` are the same as grants. This
        #       will change when admin mode is added.
        def callback(raw: Dict[str, Any]) -> None:
            raw["token_user_grants"] = raw["grants"]

        return self.get_data_secondary(
            User,
            Document,
            source=self.user,
            get_kwargs_target=get_kwargs_document,
            callback=callback,
            n=n,
        )

    def get_data_assignment_collection(
        self,
        get_kwargs_document: Dict[str, Any] = dict(other=None),
        collection: Collection | None = None,
        n: int | None = None,
    ):
        if collection is None:
            (collection,) = self.get_collections(1)

        return self.get_data_secondary(
            Collection,
            Document,
            source=collection,
            get_kwargs_target=get_kwargs_document,
            n=n,
        )

    def get_data_assignment_document(
        self,
        get_kwargs_collection: Dict[str, Any] = dict(other=None),
        document: Document | None = None,
        n: int | None = None,
    ):
        if document is None:
            (document,) = self.get_documents(1, level=Level.own)

        return self.get_data_secondary(
            Document,
            Collection,
            source=document,
            get_kwargs_target=get_kwargs_collection,
            n=n,
        )

    def get_data_secondary(
        self,
        ModelSource: ResolvableModel,
        ModelTarget: ResolvableModel,
        n: int | None = None,
        *,
        get_kwargs_source: Dict[str, Any] = dict(),
        get_kwargs_target: Dict[str, Any] = dict(),
        source: Annotated[
            Any,
            Doc("Use this to inject a source directly."),
        ] = None,
        callback: Callable[[Dict[str, Any]], None] | None = None,
    ):
        model_source = resolve_model(ModelSource)
        model_target = resolve_model(ModelTarget)
        Resolved = BaseResolved.get((model_source.__kind__, model_target.__kind__))
        assert issubclass(Resolved, BaseResolvedSecondary)

        if source is not None:
            if not isinstance(source, model_source):
                msg = f"`{source}` should be an instance of `{model_source}`."
                raise AssertionError(msg)
        else:
            (source,) = self.get(model_source, 1, **get_kwargs_source)

        n = randint(5, 15) if n is None else n
        targets = self.get(model_target, n, **get_kwargs_target)

        # NOTE: Get assocs. Assocs are always labeled by their
        model_assoc = resolve_model(Resolved.kind_assoc)  # type: ignore
        id_source_name = f"id_{Resolved._attr_name_source}"
        uuid_target_name = f"uuid_{Resolved.kind_target.name}"

        q = (
            select(model_assoc)
            .join(model_target)
            .where(
                getattr(model_assoc, id_source_name) == source.id,
                model_target.uuid.in_(uuid_target := uuids(targets)),
            )
        )
        assocs = {
            getattr(assoc, uuid_target_name): assoc for assoc in self.session.scalars(q)
        }

        # NOTE: Callback should add any additional fields. This is why this
        #       function is not often called directly.
        data = {
            Resolved._attr_name_assoc: assocs,
            Resolved._attr_name_source: source,
            Resolved._attr_name_target: targets,
            f"uuid_{Resolved._attr_name_target}": uuid_target,
        }
        if callback is not None:
            callback(data)

        return mwargs(
            Data,
            data=Resolved.model_validate(data),
        )

    @property
    def token(self) -> Token:
        # NOTE: Probably follows 7519 section 4.1,
        #       https://www.rfc-editor.org/rfc/rfc7519#section-4.1
        return mwargs(
            Token,
            sub=self.user.uuid,  # hashlib.sha256(self.user.uuid.encode()).hexdigest(),
            tier=(
                TokenPermissionTier.paid
                if not self.user.admin
                else TokenPermissionTier.admin
            ),
        )

    @property
    def token_encoded(self) -> str:
        return self.token.encode(self.auth)

    def access(self, *, method: HTTPMethod = HTTPMethod.GET) -> Access:
        return Access(self.session, self.token, method)

    def delete(
        self,
        *,
        api_origin: str,
        force: bool = True,
        method: HTTPMethod = HTTPMethod.GET,
    ) -> Delete:
        return Delete(
            self.session,
            self.token,
            method,
            api_origin=api_origin,
            access=self.access(method=method),
            force=force,
        )

    def create(
        self,
        *,
        api_origin: str,
        force: bool = True,
        method: HTTPMethod = HTTPMethod.GET,
    ) -> Create:
        return Create(
            self.session,
            self.token,
            method,
            api_origin=api_origin,
            access=self.access(method=method),
            delete=self.delete(api_origin=api_origin, force=force, method=method),
            force=force,
        )

    def requests(self, client_config, client: httpx.AsyncClient, **kwargs) -> Requests:
        # NOTE: Build configuration and context for the requests client. This
        #       is approximately that which is used by the client.
        profile = mwargs(
            ProfileConfig, token=self.token_encoded, uuid_user=self.user.uuid
        )
        context = ContextData(
            config=self.client_config_cls(
                use=UseConfig(host="default", profile="default"),
                hosts=dict(default=client_config.host),
                profiles=dict(default=profile),
            ),
            console_handler=ConsoleHandler(client_config),  # type: ignore
        )
        requests = Requests(context, client, **kwargs)
        return requests

    def dispose(self):
        logger.debug("Disposing of dummy data for user `%s`.", self.user.uuid)
        session = self.session

        session.execute(delete(User).where(User.uuid == self.user.uuid))
        session.commit()


# =========================================================================== #
# YAML Provider


# NOTE: This loads the YAML assets when this module is invoked. It might be
#       helpful to implement some sort of laziness as the IO slows things down.
class DummyYAMLProviderInfoMeta(type):
    dummies_info: ClassVar[Dict[str, "DummyProviderYAMLInfo"]] = dict()
    dummies_file: str

    def __new__(cls, name, bases, namespace):
        if name == "DummyProviderYAMLInfo":
            return super().__new__(cls, name, bases, namespace)

        if (M := namespace.get("M")) is None:
            raise ValueError("`M` must be defined.")
        elif not issubclass(M, Base):
            raise ValueError(f"`{name}.M={M}` must be a subclass of `{Base}`.")

        # Dummies. Cannot declare dummies directly.
        if namespace.get("dummies") is not None:
            raise ValueError("Cannot specify dummies explicitly.")

        if (dummies_file := namespace.get("dummies_file")) is None:
            kind = KindObject._value2member_map_[M.__tablename__]
            dummies_file = util.Path.test_assets(f"{kind.name}.yaml")
            namespace["dummies_file"] = dummies_file

        T = super().__new__(cls, name, bases, namespace)
        cls.dummies_info[T.M.__tablename__] = T  # type: ignore
        return T

    @property
    def dummies(self) -> List[Dict[str, Any]]:
        logger.debug("Loading dummy data from `%s`.", self.dummies_file)

        with open(self.dummies_file, "r") as file:
            dummies = yaml.safe_load(file)

        _msg = f"`{self.__name__}.dummies_file={self.dummies_file}`"
        if not isinstance(dummies, list):
            raise ValueError(f"{_msg} must deserialize to a list.")
        elif len(
            bad := tuple(
                index
                for index, item in enumerate(dummies)
                if not isinstance(item, dict)
            )
        ):
            raise ValueError(f"{_msg} has bad entries in positions `{bad}`.")

        return dummies


class DummyProviderYAMLInfo(metaclass=DummyYAMLProviderInfoMeta):
    # NOTE: This will matter less when the dummy data project is copmlete.
    M: ClassVar[Type[Base]]
    dummies_file: ClassVar[str]
    dummies: ClassVar[List[Dict[str, Any]]]

    @classmethod
    def preload(cls, item):
        if hasattr(item, "content"):
            if item.content is not None and not isinstance(item.content, dict):
                raise ValueError(
                    "Content should be a `dict`, not a `type(item.content)`. "
                    f"`{item = }`."
                )
        return item

    @classmethod
    def merge(cls, session: Session):
        loaded = (cls.preload(cls.M(**item)) for item in cls.dummies)
        for item in loaded:
            session.merge(item)
        session.commit()


class DummyProviderYAML(BaseDummyProvider):
    dummy_yaml_info = [
        type(
            "DummyProviderYAMLUser",
            (DummyProviderYAMLInfo,),
            dict(M=User),
        ),
        type(
            "DummyProviderYAMLCollection",
            (DummyProviderYAMLInfo,),
            dict(M=Collection),
        ),
        type(
            "DummyProviderYAMLDocument",
            (DummyProviderYAMLInfo,),
            dict(M=Document),
        ),
        type(
            "DummyProviderYAMLGrant",
            (DummyProviderYAMLInfo,),
            dict(M=AssocUserDocument),
        ),
        type(
            "DummyProviderYAMLAssignment",
            (DummyProviderYAMLInfo,),
            dict(M=AssocCollectionDocument),
        ),
        type(
            "DummyProviderYAMLEvent",
            (DummyProviderYAMLInfo,),
            dict(M=Event),
        ),
    ]

    @classmethod
    def merge(cls, session: Session) -> None:
        logger.info("Merging YAML data into database.")
        for table in Base.metadata.sorted_tables:
            yaml_data = DummyProviderYAMLInfo.dummies_info.get(table.name)
            if yaml_data is None:
                continue

            yaml_data.merge(session)


# =========================================================================== #
# Default Provider


class DummyProvider(BaseDummyProvider):
    # config: ConfigSimulatus
    session: Session
    auth: Auth

    def q_select_suitable(self):
        conds = [
            func.JSON_VALUE(User.content, "$.dummy.tainted") == false(),
            func.JSON_OVERLAPS('["YAML"]', func.JSON_VALUE(User.content, "$.tags"))
            != 1,
            User.deleted == false(),
        ]

        if self.dummy.users.maximum_uses:
            conds.append(
                func.JSON_LENGTH(User.content, "$.dummy.used_by")
                < self.dummy.users.maximum_uses
            )

        return select(User).where(*conds).order_by(func.random()).limit(1)

    def __init__(
        self,
        config: ConfigSimulatus,
        session: Session,
        *,
        auth: Auth | None = None,
        use_existing: bool | User = True,
        client_config_cls: Type | None = None,
    ):
        self.config = config
        self.dummy = config.dummy

        self.auth = auth if auth is not None else Auth.forPyTest(config)
        self.session = session
        self.client_config_cls = client_config_cls or ClientConfig

        match use_existing:
            case bool() as use_existing_bool:
                user = None
                if use_existing_bool:
                    logger.debug("Searching for existing dummy.")
                    user = session.scalar(self.q_select_suitable())

                if user is None:
                    logger.debug("Building a new dummy.")
                    self.user = self.mk_user()
                    self.mk()
                    return

                self.user = user
            case User() as user:
                logger.debug("Using existing dummy with uuid `%s`.", user.uuid)
                self.user = user
            case _:
                logger.debug("Building a new dummy.")
                self.user = self.mk_user()
                self.mk()

    def info_mark_used(self, test_fn: str) -> Self:
        logger.debug("Marking user `%s` as used by `%s`.", self.user.uuid, test_fn)
        session = self.session
        # NOTE: This should look like
        #
        #       .. code:: sql
        #
        #          UPDATE users SET info=JSON_ARRAY_APPEND(
        #            info, "$.dummy.used_by", "{test_fn}"
        #          ) WHERE uuid='000-000-000';
        #
        session.execute(
            update(User)
            .values(
                content=func.JSON_ARRAY_APPEND(User.content, "$.dummy.used_by", test_fn)
            )
            .where(User.uuid == self.user.uuid)
        )
        return self

    def info_mark_tainted(self, tainted: bool = True) -> Self:
        logger.debug("Marking user `%s` as tainted.", self.user.uuid)
        session = self.session
        session.execute(
            update(User)
            .values(content=func.JSON_REPLACE(User.content, "$.dummy.tainted", tainted))
            .where(User.uuid == self.user.uuid)
        )
        return self

    def info_is_tainted(self, maximum_use_count: int | None = None) -> bool | None:
        session = self.session
        if maximum_use_count is None:
            maximum_use_count = self.dummy.users.maximum_uses

        # NOTE: Naming easter egg.
        a = func.JSON_LENGTH(User.content, "$.dummy.used_by") >= maximum_use_count
        b = func.JSON_VALUE(User.content, "$.dummy.tainted") > 0
        q = select(or_(a, b)).where(User.uuid == self.user.uuid)
        return session.scalar(q)

    # @classmethod
    # def


# --------------------------------------------------------------------------- #


class DummyHandler:
    dummy: DummyConfig
    config: ConfigSimulatus
    sessionmaker: sqa_sessionmaker[Session]
    auth: Auth
    # user_uuids: List[str]

    def __init__(
        self,
        sessionmaker: sqa_sessionmaker,
        config: ConfigSimulatus,
        # user_uuids: List[str],
        *,
        auth: Auth | None = None,
    ):
        self.dummy = config.dummy
        self.config = config
        self.sessionmaker = sessionmaker
        self.auth = auth or Auth.forPyTest(config)
        # self.user_uuids = user_uuids

    def q_dispose(
        self,
        uuids: Set[str] | None = None,
        maximum_use_count: int | None = None,
    ):
        if maximum_use_count is None:
            maximum_use_count = self.dummy.users.maximum_uses

        conds = list()
        if uuids is not None:
            conds.append(User.uuid.in_(uuids))

        return delete(User).where(
            or_(
                func.JSON_LENGTH(User.content, "$.dummy.used_by") >= maximum_use_count,
                func.JSON_EXTRACT(User.content, "$.dummy.tainted") == true(),
            ),
            *conds,
        )

    @classmethod
    def q_select(
        cls,
        limit: int = 10,
    ):
        return (
            select(
                User.id,
                User.uuid,
                func.JSON_VALUE(User.content, "$.dummy.used_by").label("used_by"),
                func.JSON_LENGTH(User.content, "$.dummy.used_by").label(
                    "used_by_count"
                ),
                func.JSON_VALUE(User.content, "$.dummy.tainted").label("tainted"),
            )
            .where(func.JSON_CONTAINS_PATH(User.content, "all", "$.dummy.tainted"))
            .order_by(literal_column("used_by_count").desc())
            .limit(limit)
        )

    # SELECT COUNT(uuid), JSON_LENGTH(users.content, "$.dummy.used_by") AS used_by_count FROM users GROUP BY used_by_count HAVING used_by_count >= 0;
    @classmethod
    def q_select_taintedness(cls, limit: int = 10):
        _used_by_count = func.JSON_LENGTH(User.content, "$.dummy.used_by")
        used_by_count = _used_by_count.label("used_by_count")
        return (
            select(func.count(User.uuid).label("count_users"), used_by_count)
            .group_by(used_by_count)
            .order_by(used_by_count)
            .having(used_by_count >= 0)
        )

    def dispose(
        self,
        uuids: Set[str] | None = None,
        maximum_use_count: int | None = None,
        note: str | None = None,
    ):
        note = note or "Generated by `dispose`."
        with self.sessionmaker() as session:
            logger.debug("Finding and removing tainted dummies.")
            q_clean = self.q_dispose(maximum_use_count=maximum_use_count, uuids=uuids)
            util.sql(session, q_clean)
            session.execute(q_clean)
            session.commit()

    def create_report(
        self,
        note: str,
    ):
        with self.sessionmaker() as session:
            reports = ReportController(session)
            report = reports.create_aggregate(note)
            session.add(report)
            session.commit()

    def restore(
        self,
        callback: Callable[[DummyProvider, int, int], None] | None = None,
    ) -> Self:
        with self.sessionmaker() as session:
            logger.debug("Getting current user count.")
            uuids_existing = list(session.scalars(select(User.id, User.uuid)))
            n_users = len(uuids_existing)
            assert n_users is not None

            dummies: DummyConfig = self.dummy
            n_generate = dummies.users.minimum - n_users

            if n_generate < 0:
                return self

            logger.info("Generating `%s` dummies.", n_generate)
            for count in range(n_generate):
                dd = DummyProvider(
                    self.config,
                    session,
                    auth=self.auth,
                    use_existing=False,
                )
                if callback is not None:
                    callback(dd, count, n_generate)

        return self



# =========================================================================== #
import json
from typing import Any, ClassVar, Dict, Generic, Self, Tuple, TypeVar

import httpx
from fastapi import HTTPException
from pydantic import BaseModel
from rich.table import Table

# --------------------------------------------------------------------------- #
from captura import fields

# Error Message Schemas.
#
# These are important for clients that might want to use these and for testing.
# It also makes it such that all errors come from the same origin.


class ErrBase(BaseModel):
    msg: Any
    # _msg_bad_method = "Method not allowed."

    @classmethod
    def httpexception(
        cls,
        msg_name: str,
        status: int,
        *,
        dump_kwargs: Dict[str, Any] = dict(),
        **kwargs,
    ) -> HTTPException:
        if (msg := getattr(cls, msg_name, None)) is None:
            raise ValueError(f"No message named '{msg_name}'.")

        kwargs.update(msg=msg)
        self = cls(**kwargs)
        return HTTPException(status, detail=self.model_dump(mode="json", **dump_kwargs))


class ErrObjMinSchema(ErrBase):
    _msg_deleted_grant: ClassVar[str] = "Grant has been deleted."
    _msg_deleted: ClassVar[str] = "Object has been deleted."
    _msg_dne: ClassVar[str] = "Object does not exist."

    uuid_obj: fields.FieldUUID
    kind_obj: fields.FieldKindEvent | fields.FieldKindObject


class ErrAccessUser(ErrBase):
    _msg_private: ClassVar[str] = "Cannot access private user."
    _msg_modify: ClassVar[str] = "Cannot modify user."
    _msg_only_self: ClassVar[str] = "User can only access own grants."

    uuid_user: fields.FieldUUID
    uuid_user_token: fields.FieldUUID


class ErrAccessCollection(ErrBase):
    _msg_private: ClassVar[str] = "Cannot access private collection."
    _msg_modify: ClassVar[str] = "Cannot modify collection."
    _msg_homeless: ClassVar[str] = "Collection has no owner."

    uuid_user_token: fields.FieldUUID
    uuid_collection: fields.FieldUUID


# --------------------------------------------------------------------------- #


class ErrAccessEvent(ErrBase):
    _msg_not_owner: ClassVar[str] = "Cannot access event."

    uuid_event: fields.FieldUUID
    uuid_user_token: fields.FieldUUID


class ErrEventGeneral(ErrBase):
    _msg_no_root: ClassVar[str] = "Could not find parent event. Inconcievable!"

    uuid_event: fields.FieldUUID


class ErrEventKind(ErrEventGeneral):
    _msg_kind_event: ClassVar[str] = "Unexpected value for `kind_obj`."
    _msg_uuid_obj: ClassVar[str] = "Unexpected value for `uuid_obj`."

    kind: fields.FieldKindObject | fields.FieldKindEvent
    kind_expected: fields.FieldKindObject | fields.FieldKindEvent
    uuid_event: fields.FieldUUID


class ErrEventUndone(ErrEventGeneral):
    _msg_undone: ClassVar[str] = "Cannot undo undone event."

    uuid_event: fields.FieldUUID
    uuid_event_undo: fields.FieldUUID


# --------------------------------------------------------------------------- #


class ErrAccessDocumentGrantBase(ErrBase):
    _msg_dne: ClassVar[str] = "Grant does not exist."
    _msg_inconcievable: ClassVar[str] = "Inconcievable!"

    uuid_document: fields.FieldUUID
    uuid_user: fields.FieldUUID
    level_grant_required: fields.FieldLevel


class ErrAccessDocumentGrantInsufficient(ErrAccessDocumentGrantBase):
    _msg_insufficient: ClassVar[str] = "Grant insufficient."

    uuid_grant: fields.FieldUUID
    level_grant: fields.FieldLevel


class ErrAccessDocumentPending(ErrAccessDocumentGrantInsufficient):
    _msg_grant_pending: ClassVar[str] = "Grant is pending."
    _msg_grant_pending_created: ClassVar[str] = (
        "Grant is pending with `pending_from=created`."
    )

    pending_from: fields.FieldPendingFrom


class ErrAccessDocumentCannotRejectOwner(ErrBase):
    _msg_cannot_reject_owner: ClassVar[str] = (
        "Owner cannot reject grants of other owners."
    )
    uuid_user_revoker: fields.FieldUUID
    uuid_document: fields.FieldUUID
    uuid_user_revokees: fields.FieldUUIDS


class ErrUpdateGrantPendingFrom(ErrBase):
    _msg_granter: ClassVar[str] = (
        "Cannot accept grants because `pending_from` must be `granter`."
    )
    _msg_grantee: ClassVar[str] = (
        "Cannot accept grants because `pending_from` must be `grantee`."
    )

    uuid_obj: fields.FieldUUIDS
    kind_obj: fields.FieldKindObject


class ErrAssocRequestMustForce(ErrBase):
    _msg_force: ClassVar[str] = (
        "Some targets have existing assignments awaiting cleanup."
    )

    kind_target: fields.FieldKindObject
    kind_source: fields.FieldKindObject
    kind_assoc: fields.FieldKindObject

    uuid_target: fields.FieldUUIDS
    uuid_assoc: fields.FieldUUIDS
    uuid_source: fields.FieldUUID


# --------------------------------------------------------------------------- #


T_ErrDetail = TypeVar("T_ErrDetail", bound=BaseModel | str)


class ErrDetail(BaseModel, Generic[T_ErrDetail]):
    detail: T_ErrDetail

    @classmethod
    def from_response(cls, res: httpx.Response) -> Self:
        TT = cls.model_fields["detail"].annotation
        assert TT is not None
        if TT is str:
            return cls(detail=json.loads(res.content))
        return cls(detail=TT.model_validate_json(res.content))

    def compare(self, res: httpx.Response) -> Table | None:
        "Only for pytest!"

        # Assumes detail is a single layer deep.
        diff: Dict[str, Tuple[Any, Any]]
        self_from_res = self.from_response(res)
        if self.detail == self_from_res.detail:
            return None
        elif isinstance(detail := self.detail, str):
            diff = {"detail": (detail, self_from_res.detail)}
        else:
            detail_self = detail.model_dump(mode="json")  # type: ignore[union-attr]
            detail_res = self_from_res.detail.model_dump(mode="json")  # type: ignore[union-attr]
            diff = {
                exp_k: (exp_v, have_v)
                for exp_k, exp_v in detail_self.items()
                if (have_v := detail_res.get(exp_k)) is None or have_v != exp_v
            }

        table = Table(title="Error Response Discrepancy")
        table.add_column("field")
        table.add_column("value expected", style="green")
        table.add_column("value recieved", style="red")
        for key, (good, bad) in diff.items():
            table.add_row(key, good, bad)

        return table


AnyErrDetailAccessDocumentGrant = (
    ErrDetail[ErrAccessDocumentGrantBase]
    | ErrDetail[ErrAccessDocumentGrantInsufficient]
    | ErrDetail[ErrAccessDocumentPending]
)



"""Module for application configuration.

The main class defined here is `Config`. This should not `import` any internal
modules because it will result in circular imports. `Config` will be included
in controllers by dependency injection. For example, do not do

.. code:: python

    from captura.config import config

    def is_local() -> bool:
        return config.mysql.host.host


Instead do

.. code:: python

    from captura import depends

    def is_local(config: Annotated[Config, Depends(Config)]) -> bool:
        return config.m
"""

# =========================================================================== #
import enum
import secrets
from typing import Annotated, Any, ClassVar, Dict, Literal, Set, Tuple

from pydantic import BaseModel, ConfigDict, Field, SecretStr, computed_field
from sqlalchemy.engine import URL, Engine, create_engine
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from typing_extensions import Doc
from yaml_settings_pydantic import (
    BaseYamlSettings,
    YamlFileConfigDict,
    YamlSettingsConfigDict,
)

# --------------------------------------------------------------------------- #
from captura import util


class BaseHashable(BaseModel):
    """Hashable model.

    This enables caching of the configuration dependency.
    """

    hashable_fields_exclude: ClassVar[Set[str]] = set()
    model_config = ConfigDict(extra="allow")

    def __hash__(self):
        try:
            return hash(
                (type(self),)
                + tuple(
                    value
                    for key, value in self.__dict__.items()
                    if key not in self.hashable_fields_exclude
                )
            )
        except TypeError as err:
            msg = "Unhashable type encountered!"
            raise TypeError(msg) from err


class MySqlHostConfig(BaseHashable):
    """Configuration for specifying they mysql connection.

    This should include SSL once it is necessary. All field names should be
    keyword arguments to ``sqlalchemy.engine.URL.create``.

    :attr driver: The driver to use.
    :attr host: The database hostname or IP address. This defaults to the
        hostname assigned to the ``db`` container by the docker compose
        project.
    :attr port: The port on which the database exists. Defaults to ``3306``,
        the default MySQL port.
    :attr username: The login username for which connections will be
        established. This defaults to the username defined in the docker
        compose project.
    :attr password: Password corresponding to :attr:`username`.
    :attr database: The hosts database to use.
    """

    drivername_async: Annotated[str, Field("mysql+asyncmy")]
    drivername: Annotated[str, Field("mysql+pymysql")]
    host: Annotated[str, Field("db")]
    port: Annotated[int, Field(3306)]
    username: Annotated[str, Field("captura")]
    password: Annotated[SecretStr, Field("changeme")]
    database: Annotated[str, Field("captura")]


class MySqlConfig(BaseHashable):
    hashable_fields_exclude = {"engine_kwargs"}

    host: MySqlHostConfig
    engine_kwargs: Annotated[
        Dict[str, Any],
        Field(default_factory=lambda: dict()),
    ]


class Auth0ApiConfig(BaseHashable):
    audience: str | Tuple[str, ...]


class Auth0AppConfig(BaseHashable):
    client_id: str
    client_secret: SecretStr
    secret_key: Annotated[
        SecretStr, Field(default_factory=lambda: SecretStr(secrets.token_urlsafe(32)))
    ]


class Auth0Config(BaseHashable):
    """

    :attr use: Use ``auth0`` authentication when ``True``. Otherwise use a
        keypair generated by :meth:`Auth.forPyTest`. As the name indicates,
        this behaviour is desirable for tests, as using :meth:``Auth.encode``
        will enable integration testing concerning api security.
    """

    registration_code_salt: Annotated[
        bytes,
        Field(
            description=(
                "Salt for the registration code. This should be some sort of "
                "random string. See ``UserView.post_user`` and https://auth0.com/docs/customize/actions/write-your-first-action#add-a-secret"
                "about adding a secret to auth0 actions."
            )
        ),
    ]
    # registration_delay: Annotated[int, Field(description="Maximum registration code delay.")]
    issuer: Annotated[
        str,
        Field(
            description=(
                "FDQN for the auth0 tenant which will issue the tokens. This"
                " should not contain the scheme (e.g. ``https://``) or any "
                "terminating ``/``."
            )
        ),
    ]
    use: bool = True
    api: Auth0ApiConfig
    app: Auth0AppConfig

    @property
    def issuer_url(self) -> str:
        # NOTE: Standard is to not include the terminating backslash, but this
        #       is how auth0 made it so whatever.
        return f"https://{self.issuer}/"


class Environment(enum.Enum):
    production = "production"
    development = "development"


class AppConfigDev(BaseHashable):
    """Setting exclusively for development."""

    httpexc_tb: Annotated[bool, Field(default=False)]


class AppConfig(BaseHashable):
    uvicorn_port: Annotated[int, Field(default=8080)]
    uvicorn_host: Annotated[str, Field(default="0.0.0.0")]

    host_dns_name: Annotated[
        str,
        Field(default="captura.local"),
        Doc("Must not include scheme. See `host_scheme`."),
    ]
    host_scheme: Annotated[
        Literal["http://", "https://"],
        Field(default="https://"),
    ]
    host_port: Annotated[int | None, Field(default=8080)]

    environment: Annotated[
        Environment,
        Field(default=Environment.production),
    ]
    dev: Annotated[
        AppConfigDev,
        Field(default_factory=lambda: AppConfigDev.model_validate({})),
    ]
    logging_configuration_path: Annotated[
        str,
        Field(default=util.PATH_CONFIG_LOG),
    ]

    @computed_field  # type: ignore[prop-decorator]
    @property
    def is_dev(self) -> bool:
        return self.environment == Environment.development

    @computed_field  # type: ignore[prop-decorator]
    @property
    def host_url(self) -> str:
        host = f"{self.host_scheme}{self.host_dns_name}"
        if self.host_port is not None:
            host = f"{host}:{self.host_port}"
        return host


class Config(BaseHashable, BaseYamlSettings):
    model_config = YamlSettingsConfigDict(
        # NOTE: Yes, this field is labeled by the constant value computed at
        #       the initial runtime. This can be overwritten with environment
        #       changes (e.g. mocks in tests) where getting this value from the
        #       environment.
        yaml_files={
            util.PATH_CONFIG_APP: YamlFileConfigDict(
                envvar=util.prefix_env("CONFIG_APP")
            )
        },
        yaml_reload=False,
        env_prefix=util.ENV_PREFIX,
        env_nested_delimiter="__",
        extra="allow",
    )
    mysql: MySqlConfig
    auth0: Auth0Config
    app: AppConfig

    def engine(self, **engine_kwargs_extra) -> Engine:
        url = URL.create(
            **self.mysql.host.model_dump(
                exclude={"drivername", "drivername_async", "password"}
            ),
            password=self.mysql.host.password.get_secret_value(),
            drivername=self.mysql.host.drivername,
        )
        engine_kwargs = self.mysql.engine_kwargs.copy()
        engine_kwargs.update(engine_kwargs_extra)
        return create_engine(url, **engine_kwargs)

    def async_engine(self, **kwargs) -> AsyncEngine:
        url = URL.create(
            **self.mysql.host.model_dump(
                exclude={"drivername", "drivername_async"},
            ),
            drivername=self.mysql.host.drivername_async,
        )
        return create_async_engine(url, **kwargs)



import httpx
import typer

# --------------------------------------------------------------------------- #
from legere import flags
from legere.handlers import AssertionHandler
from legere.requests.assignments import DocumentAssignmentRequests
from legere.requests.base import BaseRequests, ContextData, methodize, params
from legere.requests.grants import DocumentGrantRequests


class DocumentRequests(BaseRequests):
    typer_commands = dict(
        read="req_read",
        # search="req_search",
        delete="req_delete",
        update="req_update",
        create="req_create",
    )
    typer_children = dict(
        grants=DocumentGrantRequests,
        assignments=DocumentAssignmentRequests,
    )
    grants: DocumentGrantRequests
    assignments: DocumentAssignmentRequests

    def __init__(
        self,
        context: ContextData,
        client: httpx.AsyncClient,
        *,
        handler: AssertionHandler | None = None,
        handler_methodize: bool = False,
    ):
        super().__init__(
            context,
            client,
            handler=handler,
            handler_methodize=handler_methodize,
        )
        self.grants = DocumentGrantRequests.spawn_from(self)
        self.assignments = DocumentAssignmentRequests.spawn_from(self)

    @property
    def g(self) -> DocumentGrantRequests:
        return self.grants

    @property
    def a(self) -> DocumentAssignmentRequests:
        return self.assignments

    @classmethod
    def req_delete(
        cls, _context: typer.Context, uuid_document: flags.ArgUUIDDocument
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        url = context.url(f"/documents/{uuid_document}")
        return httpx.Request("DELETE", url, headers=context.headers)

    @classmethod
    def req_create(
        cls,
        _context: typer.Context,
        *,
        name: flags.FlagName,
        description: flags.FlagDescription,
        content: flags.FlagContentOptional = None,
        public: flags.FlagPublic = False,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "POST",
            context.url("/documents"),
            json=dict(
                name=name,
                description=description,
                content=content,
                public=public,
            ),
            headers=context.headers,
        )

    @classmethod
    def req_update(
        cls,
        _context: typer.Context,
        uuid_document: flags.ArgUUIDDocument,
        *,
        name: flags.FlagNameOptional = None,
        description: flags.FlagDescriptionOptional = None,
        content: flags.FlagContentOptional = None,
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        return httpx.Request(
            "PATCH",
            context.url(f"/documents/{uuid_document}"),
            json=params(
                name=name,
                description=description,
                content=content,
            ),
            headers=context.headers,
        )

    @classmethod
    def req_read(
        cls, _context: typer.Context, uuid_document: flags.ArgUUIDDocument
    ) -> httpx.Request:
        context = ContextData.resolve(_context)
        url = context.url(f"/documents/{uuid_document}")
        return httpx.Request("GET", url, headers=context.headers)

    # @classmethod
    # def req_search(
    #     cls,
    #     _context: typer.Context,
    #     *,
    #     limit: flags.FlagLimit = 10,
    #     name_like: flags.FlagNameLike = None,
    #     description_like: flags.FlagDescriptionLike = None,
    # ):
    #     context = ContextData.resolve(_context)
    #     return httpx.Request(
    #         "GET",
    #         context.url("/documents"),
    #         params=params(
    #             limit=limit,
    #             name_like=name_like,
    #             description_like=description_like,
    #         ),
    #         headers=context.headers,
    #     )

    delete = methodize(req_delete, __func__=req_delete.__func__)  # type: ignore
    update = methodize(req_update, __func__=req_update.__func__)  # type: ignore
    create = methodize(req_create, __func__=req_create.__func__)  # type: ignore
    read = methodize(req_read, __func__=req_read.__func__)  # type: ignore
    # search = methodize(req_search, __func__=req_search.__func__)  # type: ignore


__all__ = ("DocumentRequests",)

if __name__ == "__main__":
    # --------------------------------------------------------------------------- #
    from legere.requests.base import typerize

    documents = typerize(DocumentRequests)
    documents()



# =========================================================================== #
from typing import Annotated

from fastapi import Body

# --------------------------------------------------------------------------- #
from captura.controllers.base import Data, ResolvedDocument
from captura.depends import DependsAccess, DependsCreate, DependsDelete, DependsUpdate
from captura.models import Document, Level
from captura.schemas import (
    AsOutput,
    DocumentCreateSchema,
    DocumentSchema,
    DocumentUpdateSchema,
    EventSchema,
    OutputWithEvents,
    mwargs,
)
from captura.views import args
from captura.views.base import (
    BaseView,
    OpenApiResponseCommon,
    OpenApiResponseDocumentForbidden,
    OpenApiTags,
)

OpenApiResponseDocument = {
    **OpenApiResponseCommon,
    **OpenApiResponseDocumentForbidden,
}


class DocumentView(BaseView):
    view_routes = dict(
        get_document=dict(
            url="/{uuid_document}",
            name="Get Document JSON",
            responses=OpenApiResponseDocument,
        ),
        post_document=dict(
            url="",
            name="Create Document",
        ),
        patch_document=dict(
            url="/{uuid_document}",
            name="Update Document",
            responses=OpenApiResponseDocument,
        ),
        delete_document=dict(
            url="/{uuid_document}",
            name="Delete Document and Associated Objects",
            responses=OpenApiResponseDocument,
        ),
    )
    view_router_args = dict(
        tags=[OpenApiTags.documents],
        responses=OpenApiResponseCommon,
    )

    @classmethod
    def get_document(
        cls,
        access: DependsAccess,
        uuid_document: args.PathUUIDDocument,
    ) -> AsOutput[DocumentSchema]:
        """Read a document (as `JSON`).

        To render a document, use `GET /documents/{uuid}/rendered`.
        """
        document: Document = access.document(
            uuid_document,
            level=Level.view,
            allow_public=True,
            return_data=False,
        )
        return mwargs(
            AsOutput[DocumentSchema], data=DocumentSchema.model_validate(document)
        )

    # TODO: When integration tests are written, all CUD endpoints should
    #       test that the private CUD fields are approprietly set.
    @classmethod
    def post_document(
        cls,
        create: DependsCreate,
        create_data: Annotated[DocumentCreateSchema, Body()],
    ) -> OutputWithEvents[DocumentSchema]:
        """Create a new document.

        To share your document use `POST /grants/documents/{uuid_document}` and
        to assign it to collections use
        `POST /assignments/documents/{uuid_document}`.
        """

        # TODO: Make this work as a single transaction.
        create.create_data = create_data
        data = Data[ResolvedDocument](
            token_user=create.token_user,
            event=None,
            data=ResolvedDocument.empty(token_user_grants=dict()),
            children=list(),
        )
        data_create = create.document(data)
        data_create.commit(create.session)
        (document,) = data_create.data.documents

        grant, *_ = data_create.data.token_user_grants.values()
        grant.id_document = document.id
        create.session.add(grant)
        create.session.commit()

        return mwargs(
            OutputWithEvents[DocumentSchema],
            events=[EventSchema.model_validate(data_create.event)],
            data=DocumentSchema.model_validate(document),
        )

    # TODO: Changing a document from public to private should prune the
    #       assignments from collections where the collection owner does not
    #       level of view. For now, this is not the case and they will be
    #       filtered out of results instead.
    @classmethod
    def patch_document(
        cls,
        uuid_document: args.PathUUIDDocument,
        update: DependsUpdate,
        update_data: Annotated[DocumentUpdateSchema, Body()],
    ) -> OutputWithEvents[DocumentSchema]:
        """Update document metadata."""

        update.update_data = update_data
        data = update.access.d_document(uuid_document, allow_public=False)
        update.document(data)
        data.commit(update.session)

        return mwargs(
            OutputWithEvents[DocumentSchema],
            events=[EventSchema.model_validate(data.event)],
            data=DocumentSchema.model_validate(*data.data.documents),
        )

    @classmethod
    def delete_document(
        cls,
        uuid_document: args.PathUUIDDocument,
        delete: DependsDelete,
    ) -> OutputWithEvents[DocumentSchema]:
        """Delete a **document** *(and all associated content, such as edits,
        assignments, and collections)*.

        To restore an accidental deletion (within a reasonable timeframe,
        before events and deleted objects are pruned) use
        `POST /events/objects/documents/{uuid_deleted}/restore`.

        To remove a document with no chance of restoration, use **force**. A
        record of the deletion (an `event`) will be kept for some time and
        then eventually pruned.
        """

        data = delete.a_document(uuid_document, exclude_deleted=not delete.force)
        data.commit(delete.session)
        return mwargs(
            OutputWithEvents[DocumentSchema],
            data=DocumentSchema.model_validate(data.data.documents[0]),
            events=[EventSchema.model_validate(data.event)],
        )
